"""General utilities functions."""

import re

from pathlib import Path


def camel_to_snake(name: str, sep: str = '-') -> str:
    """
    Convert camelCase or PascalCase string to snake_case.

    Parameters
    ----------
    name : str
        The string to convert.

    Returns
    -------
    str
        The converted snake_case string.
    """
    # Add underscore between lowercase and uppercase letters
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1' + sep + r'\2', name)
    # Add underscore before a sequence of uppercase letters
    snake_case = re.sub('([a-z0-9])([A-Z])', r'\1' + sep + r'\2', s1).lower()
    return snake_case


def get_absolute_path(relative_path: str) -> str:
    """Get the root path of the project."""
    return str(Path(relative_path).resolve())


# prepend stack name to the services list and stack is should be provided
# if not throw error
def prepend_stack_name(
    stack_name: str = 'default',
    services: list[str] = [],
) -> list[str]:
    """
    Prepend stack name to the services list.

    Parameters
    ----------
    stack_name : str
        The name of the stack.

    services : list[str]
        The list of services.

    Returns
    -------
    list[str]
        The list of services with the stack name prepended.
    """
    if not stack_name:
        raise ValueError('Stack name must be provided')
    return [f'{stack_name}_{service}' for service in services]



"""Profiles management screen for Sugar TUI."""

from pathlib import Path
from typing import Any, TypeVar

from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Rule,
    Static,
)

# Define the type parameter for Screen
T = TypeVar('T')


class ProfileScreen(Screen[T]):
    """Screen for managing profiles."""

    # Define CSS path for screen-specific styles
    CSS_PATH = Path(__file__).parent.parent / 'styles/screens.css'

    BINDINGS = [
        ('escape', 'app.pop_screen', 'Back'),
        ('s', "app.push_screen('services')", 'Services'),
        ('r', 'refresh', 'Refresh'),
    ]

    def compose(self) -> ComposeResult:
        """Compose the profile screen."""
        yield Header()

        yield Container(
            Vertical(
                Static('PROFILE MANAGEMENT', classes='title'),
                Horizontal(
                    Static('Manage your Sugar profiles', classes='subtitle'),
                    Static('Total Profiles: 4', classes='subtitle-right'),
                    id='screen-header',
                ),
                # Give the main table area more space
                Vertical(
                    self._create_profiles_table(), id='main-table-container'
                ),
                Horizontal(
                    Button(
                        'Add Profile', id='add-profile-btn', variant='primary'
                    ),
                    Button(
                        'Edit Profile',
                        id='edit-profile-btn',
                        variant='primary',
                    ),
                    Button(
                        'Delete Profile',
                        id='delete-profile-btn',
                        variant='error',
                    ),
                    Button('Back', id='back-btn', variant='default'),
                    id='profile-actions',
                ),
                id='screen-content',
            ),
            id='screen-container',
        )

        yield Footer()

    def _create_profiles_table(self) -> DataTable[Any]:
        """Create the profiles table."""
        # Create the table and give it explicit height styling
        profiles_table: DataTable[Any] = DataTable(id='profiles-detail-table')
        profiles_table.cursor_type = 'row'
        # Set explicit height to ensure visibility
        # profiles_table.styles.height = "1fr"
        # profiles_table.styles.min_height = 15

        profiles_table.add_columns(
            'Profile', 'Project Name', 'Config Path', 'Services', 'Status'
        )

        # Add mock data
        profiles_table.add_rows(
            [
                (
                    'development',
                    'project-dev',
                    'containers/dev/compose.yaml',
                    'frontend, api, db',
                    '● Active',
                ),
                (
                    'staging',
                    'project-staging',
                    'containers/staging/compose.yaml',
                    'All services',
                    '● Active',
                ),
                (
                    'production',
                    'project-prod',
                    'containers/prod/compose.yaml',
                    'web, api, db, cache',
                    '○ Inactive',
                ),
                (
                    'testing',
                    'project-test',
                    'containers/test/compose.yaml',
                    'test-suite, mockdb',
                    '○ Inactive',
                ),
            ]
        )

        return profiles_table

    def on_mount(self) -> None:
        """Handle mounting of the screen."""
        self.notify('Profile screen loaded')

        # Update the details section with the first profile's data
        self._update_profile_details(
            'development',
            'project-dev',
            'containers/dev/compose.yaml',
            'frontend, api, db',
        )

    def _update_profile_details(
        self, name: str, project: str, config: str, services: str
    ) -> None:
        """Update the profile details section with the selected profile."""
        # Create details container if it doesn't exist
        # Fix: Use single # for ID selector
        if not self.query('#profile-details'):
            details = Vertical(
                Static('Profile Details', classes='section-title'),
                Horizontal(
                    Label('Name:', classes='detail-label'),
                    Label(name, classes='detail-value', id='profile-name'),
                ),
                Horizontal(
                    Label('Project:', classes='detail-label'),
                    Label(project, classes='detail-value', id='project-name'),
                ),
                Horizontal(
                    Label('Config:', classes='detail-label'),
                    Label(config, classes='detail-value', id='config-path'),
                ),
                Horizontal(
                    Label('Services:', classes='detail-label'),
                    Label(
                        services, classes='detail-value', id='services-list'
                    ),
                ),
                id='profile-details',
            )

            # Add to the main container
            self.query_one('#main-table-container').mount(Rule())
            self.query_one('#main-table-container').mount(details)
        # Otherwise update existing labels
        else:
            self.query_one('#profile-name', Label).update(name)
            self.query_one('#project-name', Label).update(project)
            self.query_one('#config-path', Label).update(config)
            self.query_one('#services-list', Label).update(services)

    def action_refresh(self) -> None:
        """Refresh the profiles data."""
        self.notify('Refreshing profiles data...')

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Update details when a profile is selected in the table."""
        table = event.data_table
        row_idx = table.cursor_row
        if row_idx is not None:
            # Get the data from the selected row
            row_data = table.get_row_at(row_idx)
            # Update the details section
            self._update_profile_details(
                row_data[0], row_data[1], row_data[2], row_data[3]
            )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        button_id = event.button.id

        if button_id == 'add-profile-btn':
            self.notify('Adding new profile...')
        elif button_id == 'edit-profile-btn':
            table = self.query_one('#profiles-detail-table', DataTable)
            if table.cursor_row is not None:
                profile = table.get_row_at(table.cursor_row)[0]
                self.notify(f'Editing profile: {profile}')
            else:
                self.notify('Please select a profile first', severity='error')
        elif button_id == 'delete-profile-btn':
            table = self.query_one('#profiles-detail-table', DataTable)
            if table.cursor_row is not None:
                profile = table.get_row_at(table.cursor_row)[0]
                self.notify(f'Deleting profile: {profile}', severity='warning')
            else:
                self.notify('Please select a profile first', severity='error')
        elif button_id == 'back-btn':
            self.app.pop_screen()



"""Validation helpers for function parameters.

This module provides decorators to validate function arguments before the
wrapped function is executed.
"""

from __future__ import annotations

import functools
import inspect

from collections.abc import Collection
from typing import Any, Callable, Iterable, TypeVar

from typing_extensions import ParamSpec

from sugar.logs import SugarError, SugarLogs

P = ParamSpec('P')
R = TypeVar('R')


def validate_param(
    name: str,
    *,
    required: bool = True,
    allow_none: bool = False,
    allow_empty: bool = False,
    allow_empty_collections: bool = False,
    strip: bool = True,
    forbidden: Iterable[Any] = ('',),
    predicate: Callable[[Any], bool] | None = None,
    error_message: str | None = None,
    error_code: Any | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Validate a single function argument before running the function.

    Parameters
    ----------
    name
        Name of the argument to validate.
    required
        If True, the argument must be provided after defaults are applied.
    allow_none
        If False, ``None`` is invalid.
    allow_empty
        If False, empty strings (after optional stripping) are invalid.
    allow_empty_collections
        If False, empty collections (list/tuple/dict/set) are invalid.
    strip
        If True and the value is a string, apply ``str.strip()`` before checks.
    forbidden
        Literal values that are disallowed (checked after optional ``strip``).
    predicate
        Custom validator that returns ``True`` for valid values.
    error_message
        Custom error message; may use ``{param}`` and ``{value}`` placeholders.
    error_code
        Optional code forwarded to ``SugarLogs.raise_error``.

    Returns
    -------
    Callable
        A decorator that enforces the validation and aborts on failure.

    Raises
    ------
    ValueError
        If validation fails and no logger raises a custom error.
    """

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            bound = sig.bind_partial(*args, **kwargs)
            bound.apply_defaults()

            if name not in bound.arguments:
                _emit_error(
                    (error_message or 'Missing argument "{param}".').format(
                        param=name, value=None
                    ),
                    error_code,
                )

            value = bound.arguments[name]
            v = value.strip() if isinstance(value, str) and strip else value

            valid = True

            if required and name not in bound.arguments:
                valid = False
            elif not allow_none and v is None:
                valid = False
            elif isinstance(v, str):
                if not allow_empty and v == '':
                    valid = False
                elif v in forbidden:
                    valid = False
            elif any(v == f for f in forbidden):
                valid = False

            if (
                not allow_empty_collections
                and isinstance(v, Collection)
                and not isinstance(v, (str, bytes))
                and len(v) == 0
            ):
                valid = False

            if predicate is not None:
                try:
                    if not predicate(v):
                        valid = False
                except Exception:
                    valid = False

            if not valid:
                msg = (error_message or 'Invalid value for "{param}".').format(
                    param=name, value=value
                )
                _emit_error(msg, error_code)

            result = func(*args, **kwargs)
            return result  # mypy knows this is R from Callable[P, R]

        return wrapper

    return decorator


def require_not_blank(
    name: str,
    *,
    error_message: str = 'Value for "--{param}" is required.',
    error_code: Any | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Require that an argument is present, not None, and not blank/whitespace.

    Parameters
    ----------
    name
        Name of the argument to validate.
    error_message
        Error message; may use ``{param}`` and ``{value}`` placeholders.
    error_code
        Optional code for ``SugarLogs.raise_error``.

    Returns
    -------
    Callable
        Decorator enforcing the non-blank rule.
    """
    return validate_param(
        name,
        required=True,
        allow_none=False,
        allow_empty=False,
        allow_empty_collections=False,
        strip=True,
        forbidden=('',),
        predicate=None,
        error_message=error_message,
        error_code=error_code,
    )


def _emit_error(message: str, error_code: SugarError | None) -> None:
    """
    Emit an error via SugarLogs if available; otherwise raise ``ValueError``.

    Parameters
    ----------
    message
        Error message to report.
    error_code
        Optional code to pass to ``SugarLogs.raise_error``.
    """
    SugarLogs.raise_error(
        message, error_code or SugarError.SUGAR_INVALID_PARAMETER
    )



"""Functions for inspecting and retrieving information from containers."""

from __future__ import annotations

import re
import subprocess  # nosec B404


def get_container_name(container_id: str) -> str:
    """Get the container name for the given container_id."""
    cmd = ['docker', 'inspect', '--format={{.Name}}', container_id]
    result = subprocess.run(  # nosec B603
        cmd, capture_output=True, text=True, check=False
    )
    if not result.stdout:
        raise Exception('No container name found for the given ID')
    # Removing the leading slash from the container name
    return result.stdout.strip().lstrip('/')


def get_container_stats(container_name: str) -> tuple[float, float]:
    """
    Fetch the current memory and CPU usage of a given Docker container.

    Parameters
    ----------
        container_name (str): Name of the Docker container.

    Returns
    -------
    tuple:
        The current memory usage of the container in MB and CPU usage as
        a percentage.
    """
    command = (
        f'docker stats {container_name} --no-stream --format '
        f"'{{{{.MemUsage}}}} {{{{.CPUPerc}}}}'"
    )
    result = subprocess.run(  # nosec B602, B603
        command, capture_output=True, text=True, shell=True, check=False
    )
    output = result.stdout.strip().split()
    mem_usage_str = output[0].split('/')[0].strip()
    cpu_usage_str = output[-1].strip('%')

    mem_usage = float(re.sub(r'[^\d.]', '', mem_usage_str))
    cpu_usage = float(cpu_usage_str)

    return mem_usage, cpu_usage



"""SugarBase classes for containers."""

from __future__ import annotations

import io
import json
import os
import shlex
import sys
import tempfile

from abc import ABC, abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Any, Union

import dotenv
import sh
import yaml

from jinja2 import Environment
from jsonschema import ValidationError, validate

from sugar import __version__
from sugar.logs import SugarError, SugarLogs
from sugar.utils import camel_to_snake

TEMPLATE = Environment(
    autoescape=False,
    variable_start_string='${{',
    variable_end_string='}}',
)

SUGAR_CURRENT_PATH = Path(__file__).parent.parent


class SugarBase(ABC):
    """SugarBase defined the base structure for the Sugar classes."""

    actions: list[str] = []
    args: dict[str, str] = {}
    file: str = ''
    config: dict[str, Any] = {}
    # note: it starts with a simple command
    #       it is replaced later in the execution
    backend_app: sh.Command = sh.echo
    backend_args: list[str] = []
    defaults: dict[str, Any] = {}
    dry_run: bool = False
    env: dict[str, str] = {}
    options_args: list[str] = []
    cmd_args: list[str] = []
    service_profile: dict[str, Any] = {}
    service_names: list[str] = []
    profile_selected: str = ''
    verbose: bool = False
    hooks: dict[str, list[dict[str, Any]]] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the actions list for all the created commands."""
        super().__init_subclass__(**kwargs)
        # Ensure each subclass has its own actions list
        cls.actions = cls.actions.copy()
        prefix = '_cmd_'
        prefix_len = len(prefix)
        for name, value in cls.__dict__.items():
            if callable(value) and name.startswith(prefix):
                action_name = name[prefix_len:]
                cls.actions.append(action_name)

    def __init__(self) -> None:
        """Initialize SugarBase instance."""
        self.file = '.sugar.yaml'
        self.verbose = False
        self.dry_run = False
        self.args: dict[str, str] = {}
        self.options_args: list[str] = []
        self.cmd_args: list[str] = []
        self.config: dict[str, Any] = {}
        self.backend_args: list[str] = []
        self.defaults: dict[str, Any] = {}
        self.env: dict[str, str] = {}
        self.service_profile: dict[str, Any] = {}
        self.service_names: list[str] = []
        self.profile_selected: str = ''

    def _setup_load(self, **kwargs: Any) -> None:
        """Set up the configuration for running the commands."""
        args: dict[str, str] = kwargs.get('args', {})
        options_args: list[str] = kwargs.get('options_args', [])
        cmd_args: list[str] = kwargs.get('cmd_args', [])

        self.args = deepcopy(args)
        self.options_args = deepcopy(options_args)
        # note: it needs to be refactored
        self.cmd_args = deepcopy(cmd_args)

        self._load_config()
        self._load_env()
        self._load_defaults()
        self._load_root_services()

        self._load_profile()
        self._load_backend()

        self._verify_config()

    @abstractmethod
    def _load_backend(self) -> None:
        """
        Initialize the backend application and its arguments.

        This method must be implemented by subclasses to set:
        - self.backend_app
        - self.backend_args
        """
        pass

    def _call_backend_app(
        self,
        action: str,
        services: list[str] = [],
        nodes: list[str] = [],
        options_args: list[str] = [],
        cmd_args: list[str] = [],
        _out: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stdout,
        _err: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stderr,
    ) -> None:
        # validation
        if services and nodes:
            SugarLogs.raise_error(
                'Give services or nodes arguments, not both.',
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        nodes_or_services = services or nodes

        # Execute pre-run hooks
        extension = camel_to_snake(
            self.__class__.__name__.replace('Sugar', '')
        )
        self._execute_hooks('pre-run', extension, action)

        sh_extras = {
            '_in': sys.stdin,
            '_out': _out,
            '_err': _err,
            '_no_err': True,
            '_env': os.environ,
            '_bg': True,
            '_bg_exc': False,
        }

        positional_parameters = [
            *self.backend_args,
            *[action],
            *options_args,
            *nodes_or_services,
            *cmd_args,
        ]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> {self.backend_app} {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        p = self.backend_app(
            *positional_parameters,
            **sh_extras,
        )

        try:
            p.wait()
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            pid = p.pid
            p.kill()
            SugarLogs.raise_error(
                f'Process {pid} killed.', SugarError.SH_KEYBOARD_INTERRUPT
            )
        self._execute_hooks('post-run', extension, action)

    def _check_config_file(self) -> bool:
        return Path(self.file).exists()

    # Check if services item is given
    def _check_services_item(self) -> bool:
        return hasattr(self.config, 'services')

    def _execute_hooks(
        self, hook_type: str, extension: str, action: str
    ) -> None:
        """Execute hooks specific type, extension, and action."""
        hooks = self.hooks.get(hook_type, [])

        sh_extras = {
            '_in': sys.stdin,
            '_out': sys.stdout,
            '_err': sys.stderr,
            '_no_err': True,
            '_env': os.environ,
        }

        fd, filepath = tempfile.mkstemp(suffix='sugar', text=True)

        for hook in hooks:
            targets = hook.get('targets', {})
            target_ext = targets.get(extension, [])
            hook_name = hook.get('name', '')

            if not target_ext or action not in target_ext:
                continue

            SugarLogs.print_info(f'Running {hook_type} hook: {hook_name} ...')
            cmd = hook.get('run', '').strip()

            with open(filepath, 'w') as f:
                f.write(cmd)

            sh.xonsh(filepath, **sh_extras)

        os.remove(filepath)

    # set default profile main
    def _load_root_services(self) -> None:
        """Load services attribute in the root of the configuration."""
        # must set the default profile
        services = self.config.get('services', {})

        if not services:
            return

        self.config['profiles'] = {
            'main': {
                'project-name': services.get('project-name'),
                'config-path': services.get('config-path'),
                'env-file': services.get('env-file'),
                'services': {
                    'default': services.get('default', []),
                    'available': services.get('available', []),
                },
            }
        }
        self.defaults['profile'] = 'main'
        self.profile_selected = 'main'
        self.service_profile = deepcopy(self.config['profiles']['main'])
        del self.config['services']

    def _filter_service_profile(self) -> None:
        profiles = self.config['profiles']

        # Verify if project-name is not null
        default_project_name = self.defaults.get('project-name', '') or ''

        for profile_name, profile_data in profiles.items():
            if profile_name == self.profile_selected:
                if default_project_name and 'project-name' not in profile_data:
                    # just use default value if "project-name" is not set
                    profile_data['project-name'] = default_project_name
                if not profile_data.get('services', {}).get('default'):
                    # if default is not given or it is empty or null,
                    # use as default all the services available
                    default_services = [
                        service['name']
                        for service in profile_data.get('services', {}).get(
                            'available', []
                        )
                    ]
                    profile_data['services']['default'] = default_services
                self.service_profile = profile_data
                return

        SugarLogs.raise_error(
            f'The given profile service "{self.profile_selected}" '
            'was not found in the configuration file.',
            SugarError.SUGAR_MISSING_PARAMETER,
        )

    def _load_config(self) -> None:
        with open(self.file, 'r') as f:
            # escape template tags
            content = f.read()
            f_content = io.StringIO(content)
            self.config = yaml.safe_load(f_content)

        # check if either  services or  profiles are present
        if not (self.config.get('services') or self.config.get('profiles')):
            SugarLogs.raise_error(
                'Either `services` OR  `profiles` flag must be given',
                SugarError.SUGAR_INVALID_CONFIGURATION,
            )
        # check if both services and profiles are present
        if self.config.get('services') and self.config.get('profiles'):
            SugarLogs.raise_error(
                '`services` and `profiles` flags given, only 1 is allowed.',
                SugarError.SUGAR_INVALID_CONFIGURATION,
            )

        self._validate_config()

        # Load hooks
        self.hooks = self.config.get('hooks', {})

    def _load_defaults(self) -> None:
        _defaults = self.config.get('defaults', {})

        for k, v in _defaults.items():
            unescaped_value = v if isinstance(v, str) else str(v)

            _defaults[k] = yaml.safe_load(
                TEMPLATE.from_string(unescaped_value).render(env=self.env)
            )

        self.defaults = _defaults

    def _load_env(self) -> None:
        self.env = dict(os.environ)

        env_files = self.config.get('env-file', '')

        if isinstance(env_files, str):
            env_files = [env_files] if env_files else []

        if not env_files:
            return

        for env_file in env_files:
            if not env_file:
                continue

            env_file_path = env_file

            if not env_file.startswith('/'):
                # use .sugar file as reference for the working
                # directory for the .env file
                env_file_path = str(Path(self.file).parent / env_file)

            if not Path(env_file_path).exists():
                SugarLogs.raise_error(
                    f'Env file was not found: {env_file_path}',
                    SugarError.SUGAR_INVALID_CONFIGURATION,
                )
            self.env.update(dotenv.dotenv_values(env_file_path))  # type: ignore

    def _load_profile(self) -> None:
        if not self.profile_selected:
            default_profile = self.defaults.get('profile', '')
            if not default_profile:
                SugarLogs.raise_error(
                    'The service profile parameter or default '
                    "profile configuration weren't defined.",
                    SugarError.SUGAR_INVALID_PARAMETER,
                )
            self.profile_selected = default_profile
        self._filter_service_profile()

    def _get_list_args(self, args: str) -> list[str]:
        """Return a list with the name of the service if any."""
        if not args.strip():
            return []

        return shlex.split(args)

    def _get_service_name(self, service: str) -> list[str]:
        """Return a list with the name of the service if any."""
        return [service] if service else []

    def _get_services_names(self, **kwargs: Any) -> list[str]:
        if 'all' not in kwargs and 'services' not in kwargs:
            # the command doesn't specify services (e.g. version)
            return []

        _arg_services = kwargs.get('services', '')
        _arg_all = kwargs.get('all', False)

        services_config = self.service_profile['services']
        service_names: list[str] = []
        services_default = services_config.get('default', [])

        if _arg_all:
            service_names = [
                service['name']
                for service in self.service_profile.get('services', {}).get(
                    'available', []
                )
            ]
        elif _arg_services:
            service_names = _arg_services.split(',')
        elif services_default:
            service_names = services_default
        else:
            SugarLogs.raise_error(
                'If you want to execute the operation for all services, '
                'use --all parameter.',
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        return service_names

    def _validate_config(self) -> None:
        """
        Validate the .sugar.yaml against the predefined JSON Schema.

        Raises
        ------
            SugarError: If the configuration does not conform to the schema.
        """
        try:
            with open(SUGAR_CURRENT_PATH / 'schema.json', 'r') as schema_file:
                schema = json.load(schema_file)

            config_data = self.config

            # Validate the configuration against the schema
            validate(instance=config_data, schema=schema)

            if self.verbose:
                SugarLogs.print_info('Configuration validation successful.')

        except ValidationError as ve:
            error_message = f'Configuration validation error: {ve.message}'
            SugarLogs.raise_error(
                error_message, SugarError.CONFIG_VALIDATION_ERROR
            )
        except yaml.YAMLError as ye:
            error_message = f'YAML parsing error: {ye}'
            SugarLogs.raise_error(error_message, SugarError.YAML_PARSING_ERROR)
        except json.JSONDecodeError as je:
            error_message = f'JSON schema decoding error: {je}'
            SugarLogs.raise_error(
                error_message, SugarError.JSON_SCHEMA_DECODING_ERROR
            )
        except FileNotFoundError:
            error_message = f'Configuration file {self.file} not found.'
            SugarLogs.raise_error(
                error_message, SugarError.SUGAR_CONFIG_FILE_NOT_FOUND
            )
        except Exception as e:
            error_message = (
                f'Unexpected error during configuration validation: {e}'
            )
            SugarLogs.raise_error(
                error_message, SugarError.CONFIG_VALIDATION_UNEXPECTED_ERROR
            )

    def _verify_config(self) -> None:
        if not self._check_config_file():
            SugarLogs.raise_error(
                'Config file .sugar.yaml not found.',
                SugarError.SUGAR_INVALID_CONFIGURATION,
            )

        if not len(self.config.get('profiles', {})):
            SugarLogs.raise_error(
                'No service profiles found.',
                SugarError.SUGAR_INVALID_CONFIGURATION,
            )

    def _version(self) -> None:
        SugarLogs.print_info(f'Sugar Version: {__version__}')

    def load(
        self,
        file: str = '.sugar.yaml',
        profile: str = '',
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        """Load sugar configuration."""
        self.file = file
        self.profile_selected = profile
        self.dry_run = dry_run
        self.verbose = verbose

        self._setup_load()



"""Services management screen for interacting with running containers."""

from pathlib import Path
from typing import Any, TypeVar

from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Rule,
    Static,
)

T = TypeVar('T')


class ServiceScreen(Screen[T]):
    """Screen to display and manage services."""

    CSS_PATH = Path(__file__).parent.parent / 'styles/screens.css'

    BINDINGS = [
        ('escape', 'app.pop_screen', 'Back'),
        ('l', "app.push_screen('logs')", 'Logs'),
        ('d', "app.push_screen('details')", 'Details'),
        ('r', 'refresh', 'Refresh'),
    ]

    def compose(self) -> ComposeResult:
        """Compose the service screen layout."""
        yield Header()

        yield Container(
            Vertical(
                Static('SERVICE MANAGEMENT', classes='title'),
                Horizontal(
                    Static(
                        'Manage your container services', classes='subtitle'
                    ),
                    Static('Total Services: 6', classes='subtitle-right'),
                    id='screen-header',
                ),
                # Give the main table area more space
                Vertical(
                    self._create_services_table(), id='main-table-container'
                ),
                Horizontal(
                    Button('Start', id='start-service', variant='success'),
                    Button('Stop', id='stop-service', variant='error'),
                    Button('Restart', id='restart-service', variant='warning'),
                    Button('Logs', id='logs-service', variant='primary'),
                    Button('Details', id='details-service', variant='default'),
                    Button('Back', id='back-btn', variant='default'),
                    id='profile-actions',
                ),
                id='screen-content',
            ),
            id='screen-container',
        )

        yield Footer()

    def _create_services_table(self) -> DataTable[Any]:
        """Create the services table."""
        # Create table with explicit height styling
        table: DataTable[Any] = DataTable(id='services-table')
        table.cursor_type = 'row'

        # Apply the same height constraints as profiles table
        table.styles.height = 7
        table.styles.max_height = 7

        table.add_columns(
            'Service', 'Status', 'Ports', 'Image', 'Container ID'
        )

        # This would be populated from actual Sugar config in the future
        table.add_rows(
            [
                (
                    'frontend',
                    '● Running',
                    '3000:3000',
                    'node:16',
                    'abc123def456',
                ),
                (
                    'api',
                    '● Running',
                    '8000:8000',
                    'python:3.10',
                    'def456ghi789',
                ),
                (
                    'db',
                    '● Running',
                    '5432:5432',
                    'postgres:14',
                    '789ghi101112',
                ),
                (
                    'redis',
                    '● Running',
                    '6379:6379',
                    'redis:latest',
                    'jkl131415mno',
                ),
                (
                    'nginx',
                    '● Running',
                    '80:80, 443:443',
                    'nginx:latest',
                    'pqr161718stu',
                ),
                ('worker', '○ Stopped', '-', 'python:3.10', ''),
            ]
        )

        return table

    def on_mount(self) -> None:
        """Handle mounting of the screen."""
        self.notify('Services screen loaded')

        self._update_service_details(
            'frontend', '● Running', '3000:3000', 'node:16', 'abc123def456'
        )

    def _update_service_details(
        self, name: str, status: str, ports: str, image: str, container_id: str
    ) -> None:
        """Update the service details section with the selected service."""
        # Create details container if it doesn't exist
        if not self.query('#service-details'):
            details = Vertical(
                Static('Service Details', classes='section-title'),
                Horizontal(
                    Label('Name:', classes='detail-label'),
                    Label(name, classes='detail-value', id='service-name'),
                ),
                Horizontal(
                    Label('Status:', classes='detail-label'),
                    Label(status, classes='detail-value', id='service-status'),
                ),
                Horizontal(
                    Label('Ports:', classes='detail-label'),
                    Label(ports, classes='detail-value', id='service-ports'),
                ),
                Horizontal(
                    Label('Image:', classes='detail-label'),
                    Label(image, classes='detail-value', id='service-image'),
                ),
                Horizontal(
                    Label('Container ID:', classes='detail-label'),
                    Label(
                        container_id, classes='detail-value', id='container-id'
                    ),
                ),
                id='service-details',
            )

            # Add to the main container
            self.query_one('#main-table-container').mount(Rule())
            self.query_one('#main-table-container').mount(details)
        # Otherwise update existing labels
        else:
            self.query_one('#service-name', Label).update(name)
            self.query_one('#service-status', Label).update(status)
            self.query_one('#service-ports', Label).update(ports)
            self.query_one('#service-image', Label).update(image)
            self.query_one('#container-id', Label).update(container_id)

    def action_refresh(self) -> None:
        """Refresh the services data."""
        self.notify('Refreshing services data...')

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Update details when a service is selected in the table."""
        table = event.data_table
        row_idx = table.cursor_row
        if row_idx is not None:
            # Get the data from the selected row
            row_data = table.get_row_at(row_idx)
            # Update the details section
            self._update_service_details(
                row_data[0], row_data[1], row_data[2], row_data[3], row_data[4]
            )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        button_id = event.button.id

        if button_id == 'back-btn':
            self.app.pop_screen()
            return

        table = self.query_one('#services-table', DataTable)
        if table.cursor_row is None:
            self.notify('Please select a service first', severity='error')
            return

        service = table.get_row_at(table.cursor_row)[0]

        if button_id == 'start-service':
            self.notify(f'Starting {service}...', title='Starting')
        elif button_id == 'stop-service':
            self.notify(f'Stopping {service}...', title='Stopping')
        elif button_id == 'restart-service':
            self.notify(f'Restarting {service}...', title='Restarting')
        elif button_id == 'logs-service':
            try:
                self.app.push_screen('logs')
            except Exception as e:
                self.notify(f'Error opening logs: {e!s}', severity='error')
        elif button_id == 'details-service':
            try:
                self.app.push_screen('details')
            except Exception as e:
                self.notify(f'Error opening details: {e!s}', severity='error')



"""TUI extension for Sugar."""

from __future__ import annotations

from sugar.docs import docparams
from sugar.extensions.base import SugarBase
from sugar.logs import SugarLogs


class SugarTUI(SugarBase):
    """Terminal User Interface extension for Sugar."""

    def _load_backend(self) -> None:
        """Load backend and backend parameters."""
        pass

    @docparams({})  # Empty dict since we don't have parameters
    def _cmd_tui(self) -> None:
        """Launch the Sugar Terminal User Interface."""
        SugarLogs.print_info('Starting Sugar TUI...')
        try:
            from sugar.tui.app import run

            SugarLogs.print_info('Successfully imported run function')
            run()
            SugarLogs.print_info(
                'TUI has completed'
            )  # This will only print if run() completes/exits
        except Exception as e:
            SugarLogs.print_warning(f'Error launching TUI: {e}')
            import traceback

            traceback.print_exc()



"""
Sugar plugin for docker swarm and docker service.

Swarm Commands:
  - config      Manage Swarm configs
  - node        Manage Swarm nodes
  - secret      Manage Swarm secrets
  - service     Manage Swarm services
  - stack       Manage Swarm stacks
  - swarm       Manage Swarm
"""

from __future__ import annotations

import io
import sys

from typing import Any, Union

import sh

from sugar.docs import docparams
from sugar.extensions.compose import SugarComposeBase
from sugar.logs import SugarError, SugarLogs
from sugar.utils import prepend_stack_name
from sugar.validation import require_not_blank

MSG_ERROR_STACK_NAME = 'Stack name must be provided'
MSG_ERROR_NODES_NAME = 'Node name(s) must be provided'

doc_profile = {
    'profile': 'Specify the profile name of the services you want to use.'
}
doc_options = {
    'options': (
        'Specify the options for the backend command. '
        'E.g.: `--  --advertise-addr 192.168.1.1`.'
    )
}
doc_service = {'service': 'Set the service for the swarm command.'}
doc_services = {
    'services': 'Set the services separated by comma for the swarm command.'
}
doc_node = {'node': 'Set the node for the swarm command.'}
doc_nodes = {
    'nodes': 'Set the nodes separated by comma for the swarm command.'
}
doc_all_services = {'all': 'Use all services for the command.'}
doc_all_nodes = {'all': 'Use all nodes for the command.'}
doc_subcommand = {'subcommand': 'Subcommand to execute for the node command.'}
doc_stack = {'stack': 'Name of the stack to deploy'}
doc_compose_file = {
    'file': 'Path to a Compose file (overrides the one from group/profile)',
}

doc_common_no_services = {**doc_profile, **doc_options}
doc_common_service = {**doc_profile, **doc_service, **doc_options}
doc_common_services = {
    **doc_profile,
    **doc_services,
    **doc_all_services,
    **doc_options,
}

doc_common_services_stack = {**doc_common_services, **doc_stack}

doc_stack_plus_options = {**doc_stack, **doc_options}

doc_common_node = {**doc_profile, **doc_node, **doc_options}
doc_common_nodes = {**doc_profile, **doc_nodes, **doc_all_nodes, **doc_options}
doc_node_command = {
    **doc_profile,
    **doc_options,
    **doc_nodes,
    **doc_all_nodes,
    **doc_subcommand,
}

# Node command specific docs
doc_node_options = {
    'demote': 'Demote one or more nodes from manager in the swarm',
    'inspect': 'Display detailed information on one or more nodes',
    'ls': 'List nodes in the swarm',
    'promote': 'Promote one or more nodes to manager in the swarm',
    'ps': 'List tasks running on one or more nodes, defaults to current node',
    'rm': 'Remove one or more nodes from the swarm',
    'update': 'Update a node',
}

doc_service_options = {
    'create': 'Create a new service',
    'inspect': 'Display detailed information on one or more services',
    'logs': 'Fetch the logs of a service or task',
    'ls': 'List services',
    'ps': 'List the tasks of one or more services',
    'rm': 'Remove one or more services',
    'rollback': "Revert changes to a service's configuration",
    'scale': 'Scale one or multiple replicated services',
    'update': 'Update a service',
}

# Service command specific docs
doc_services_logs_options = {
    'details': 'Show extra details provided to logs',
    'stack': 'Name of the stack to inspect',
    'follow': 'Follow log output',
    'no_resolve': 'Do not map IDs to Names in output',
    'no_task_ids': 'Do not include task IDs in output',
    'no_trunc': 'Do not truncate output',
    'raw': 'Do not neatly format logs',
    'since': 'Show logs since timestamp or relative time (e.g. 42m)',
    'tail': 'Number of lines to show from the end of the logs (default all)',
    'timestamps': 'Show timestamps',
}
doc_service_actions = {
    'create': 'Create a new service (use --  for all parameters)',
    'inspect': 'Display detailed information on one or more services',
    'logs': 'Fetch the logs of a service or task',
    'ls': 'List services',
    'ps': 'List the tasks of one or more services',
    'rm': 'Remove one or more services',
    'rollback': "Revert changes to a service's configuration",
    'scale': 'Scale one or multiple replicated services',
    'update': 'Update a service',
}
doc_rollback_options = {
    'detach': (
        'Exit immediately instead of waiting for the service to converge'
    ),
    'quiet': 'Suppress progress output',
}
doc_scale_options = {
    'detach': (
        'Exit immediately instead of waiting for the service to converge'
    ),
    'replicas': (
        'Number of replicas per service '
        '(comma-separated list of service=replicas pairs)'
    ),
}
doc_update_options = {
    'detach': (
        'Exit immediately instead of waiting for the service to converge'
    ),
    'quiet': 'Suppress progress output',
    'image': 'Service image tag',
    'replicas': 'Number of tasks',
    'force': 'Force update even if no changes require it',
    'rollback': 'Rollback to previous specification',
    'env_add': 'Add/update env vars (comma-separated NAME=VALUE list)',
    'label_add': 'Add/update service labels (comma-separated key=value list)',
}


class SugarSwarmBase(SugarComposeBase):
    """SugarSwarmBase provides base functionalities."""

    def _load_backend(self) -> None:
        """Load the backend for the swarm commands."""
        self._load_backend_app()
        self._load_backend_args()

    def _load_backend_app(self) -> None:
        self.backend_app = sh.docker

    def _load_backend_args(self) -> None:
        self.backend_args = []

    def _get_services_names(self, **kwargs: Any) -> list[str]:
        """
        Override to handle swarm service names without requiring config.

        For swarm commands, services are specified directly on the command line
        and don't need to be defined in a config file.
        """
        if 'all' not in kwargs and 'services' not in kwargs:
            # The command doesn't specify services
            return []

        _arg_services = kwargs.get('services', '')
        _arg_stack = kwargs.get('stack', '')

        # For swarm, we don't use the 'all' flag, only explicit services
        if not _arg_services:
            SugarLogs.raise_error(
                'Service name must be provided for this command '
                '(use --services service1,service2)',
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        # Simply split the comma-separated service names
        services_list: list[str] = _arg_services.split(',')

        if _arg_stack:
            services_list = [f'{_arg_stack}_{name}' for name in services_list]

        return services_list

    def _call_stack_command(
        self,
        stack_name: str = '',
        options_args: list[str] = [],
        backend_args: list[str] = [],
        include_compose_files: bool = False,
        _out: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stdout,
        _err: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stderr,
    ) -> None:
        """Call docker stack commands with proper structure."""
        # Build the full command: stack deploy -c file stackname

        # Check if compose file should be included
        self.backend_args = backend_args.copy()

        if include_compose_files:
            for config_file in self._get_config_files_path():
                self.backend_args.extend(['-c', config_file])

        # Call with the stack name as the main command/argument
        self._call_backend_app(
            stack_name,
            options_args=options_args,
            _out=_out,
            _err=_err,
        )

    def _call_command(
        self,
        subcommand: str,
        services: list[str] = [],
        nodes: list[str] = [],
        options_args: list[str] = [],
        cmd_args: list[str] = [],
        _out: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stdout,
        _err: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stderr,
    ) -> None:
        """Call docker swarm commands with proper structure."""
        if services and nodes:
            SugarLogs.raise_error(
                'Give services or nodes arguments, not both.',
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        nodes_or_services: dict[str, list[str]] = {'services': []}
        if services:
            nodes_or_services = {'services': services}
        elif nodes:
            nodes_or_services = {'nodes': nodes}

        self._call_backend_app(
            subcommand,
            options_args=options_args,
            cmd_args=cmd_args,
            _out=_out,
            _err=_err,
            **nodes_or_services,
        )

    def _get_services_from_stack(self, stack: str) -> list[str]:
        """Get all services from a stack."""
        try:
            output = io.StringIO()
            self.backend_app(
                'stack',
                'services',
                stack,
                '--format',
                '{{.Name}}',
                _out=output,
            )
            services_output = output.getvalue()

            services = [
                service
                for service in services_output.strip().split('\n')
                if service
            ]
            if not services:
                SugarLogs.raise_error(
                    f'No services found in stack {stack}',
                    SugarError.SUGAR_INVALID_PARAMETER,
                )
            return services
        except Exception as e:
            SugarLogs.raise_error(
                f'Failed to get services from stack {stack}: {e!s}',
                SugarError.SUGAR_COMMAND_ERROR,
            )
            return []


class SugarSwarm(SugarSwarmBase):
    """
    SugarSwarm provides the docker swarm commands.

    Commands: ca, init, join, join, leave, unlock, unlock, update.
    """

    def _load_backend_args(self) -> None:
        self.backend_args = ['swarm']

    @docparams(doc_common_no_services)
    def _cmd_ca(
        self,
        options: str = '',
    ) -> None:
        """Display and rotate the root CA."""
        options_args = self._get_list_args(options)
        self._call_command('join', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_init(
        self,
        options: str = '',
    ) -> None:
        """Initialize a swarm.

        This command initializes a new swarm on the current Docker engine.
        """
        # For swarm init, use the wrapper method instead
        options_args = self._get_list_args(options)
        self._call_command('init', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_join(
        self,
        options: str = '',
    ) -> None:
        """Join a swarm as a node and/or manager."""
        options_args = self._get_list_args(options)
        self._call_command('join', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_join_token(
        self,
        options: str = '',
    ) -> None:
        """Manage join tokens."""
        options_args = self._get_list_args(options)
        self._call_command('join-token', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_leave(
        self,
        options: str = '',
    ) -> None:
        """Leave the swarm."""
        options_args = self._get_list_args(options)
        self._call_command('leave', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_unlock(
        self,
        options: str = '',
    ) -> None:
        """Unlock swarm."""
        options_args = self._get_list_args(options)
        self._call_command('unlock', options_args=options_args)

    @docparams(doc_common_no_services)
    def _cmd_unlock_key(
        self,
        options: str = '',
    ) -> None:
        """Manage the unlock key."""
        options_args = self._get_list_args(options)
        self._call_command('unlock-key', options_args=options_args)

    @docparams({**doc_update_options})
    def _cmd_update(
        self,
        detach: bool = False,
        quiet: bool = False,
        image: str = '',
        replicas: str = '',
        force: bool = False,
        rollback: bool = False,
        env_add: str = '',
        label_add: str = '',
        options: str = '',
    ) -> None:
        """Update services (docker service update)."""
        names: list[str] = []
        opts = self._get_list_args(options)

        if detach:
            opts.append('--detach')
        if quiet:
            opts.append('--quiet')
        if force:
            opts.append('--force')
        if rollback:
            opts.append('--rollback')
        if image:
            opts.extend(['--image', image])
        if replicas:
            opts.extend(['--replicas', replicas])

        if env_add:
            for pair in env_add.split(','):
                if pair.strip():
                    opts.extend(['--env-add', pair.strip()])
        if label_add:
            for pair in label_add.split(','):
                if pair.strip():
                    opts.extend(['--label-add', pair.strip()])

        self._call_command('update', services=names, options_args=opts)


class SugarSwarmService(SugarSwarmBase):
    """
    SugarSwarmService provides the `docker service` commands.

    Commands: create, inspect, logs, ls, ps, rm, rollback, scale, update.
    """

    def _load_backend_args(self) -> None:
        self.backend_args = ['service']

    def _perform_service_rollback(
        self, service: str, options_args: list[str]
    ) -> bool:
        """Perform rollback for a single service; True on success."""
        try:
            output = io.StringIO()
            error = io.StringIO()
            self.backend_app(
                'service',
                'rollback',
                *options_args,
                service,
                _out=output,
                _err=error,
                _ok_code=[0, 1],
            )
            err = error.getvalue()
            if 'does not have a previous spec' in err:
                SugarLogs.print_warning(
                    f'Service {service} has no previous version to rollback to'
                )
                return False
            if err:
                SugarLogs.print_warning(
                    f'Failed to rollback service {service}: {err.strip()}'
                )
                return False
            print(f'Successfully rolled back service {service}')
            return True
        except Exception as e:
            SugarLogs.print_warning(
                f'Error rolling back service {service}: {e!s}'
            )
            return False

    # ---------- service commands ----------
    @docparams(doc_options)
    def _cmd_create(self, options: str = '') -> None:
        """Create a new service (docker service create)."""
        if not options:
            SugarLogs.raise_error(
                'Options must be provided for "create". '
                'Include --name, image, etc. inside -- .',
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        self._call_command(
            'create', services=[], options_args=self._get_list_args(options)
        )

    @require_not_blank('stack')
    @docparams({**doc_common_services_stack})
    def _cmd_inspect(
        self,
        stack: str = '',
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Inspect one or more services (docker service inspect)."""
        services_names = self._get_services_names(
            services=services, all=all, stack=stack
        )
        self._call_command(
            'inspect',
            services=services_names,
            options_args=self._get_list_args(options),
        )

    @docparams({**doc_common_services, **doc_services_logs_options})
    def _cmd_logs(
        self,
        services: str = '',
        all: bool = False,
        stack: str = '',
        details: bool = False,
        follow: bool = False,
        no_resolve: bool = False,
        no_task_ids: bool = False,
        no_trunc: bool = False,
        raw: bool = False,
        since: str = '',
        tail: str = '',
        timestamps: bool = False,
        options: str = '',
    ) -> None:
        """Fetch logs of a service or task (docker service logs)."""
        svc_names = prepend_stack_name(
            stack_name=stack,
            services=self._get_services_names(services=services, all=all),
        )
        opts = self._get_list_args(options)
        if details:
            opts.append('--details')
        if follow:
            opts.append('--follow')
        if no_resolve:
            opts.append('--no-resolve')
        if no_task_ids:
            opts.append('--no-task-ids')
        if no_trunc:
            opts.append('--no-trunc')
        if raw:
            opts.append('--raw')
        if timestamps:
            opts.append('--timestamps')
        if since:
            opts.extend(['--since', since])
        if tail:
            opts.extend(['--tail', tail])

        self._call_command('logs', services=svc_names, options_args=opts)

    @docparams(doc_common_no_services)
    def _cmd_ls(self, options: str = '') -> None:
        """List services (docker service ls)."""
        self._call_command('ls', options_args=self._get_list_args(options))

    @docparams(doc_common_services)
    def _cmd_ps(
        self, services: str = '', all: bool = False, options: str = ''
    ) -> None:
        """List tasks of services (docker service ps)."""
        names = self._get_services_names(services=services, all=all)
        self._call_command(
            'ps', services=names, options_args=self._get_list_args(options)
        )

    @docparams(doc_common_services)
    def _cmd_rm(
        self, services: str = '', all: bool = False, options: str = ''
    ) -> None:
        """Remove services (docker service rm)."""
        names = self._get_services_names(services=services, all=all)
        self._call_command(
            'rm', services=names, options_args=self._get_list_args(options)
        )

    @docparams({**doc_common_service, **doc_rollback_options, **doc_stack})
    def _cmd_rollback(
        self,
        service: str = '',
        stack: str = '',
        detach: bool = False,
        quiet: bool = False,
        options: str = '',
    ) -> None:
        """
        Roll back services to previous versions (docker service rollback).

        Supports either explicit --services or whole --stack.
        """
        opts = self._get_list_args(options)
        if detach:
            opts.append('--detach')
        if quiet:
            opts.append('--quiet')

        # Determine targets
        targets = self._get_services_names(
            services=service,
            all=False,
            stack=stack,
        )

        if not targets:
            SugarLogs.print_warning('No services specified for rollback')
            return

        ok, bad = 0, 0
        for svc in targets:
            if self._perform_service_rollback(svc, opts):
                ok += 1
            else:
                bad += 1
        print(f'Rollback complete: {ok} succeeded, {bad} failed')

    @require_not_blank(name='replicas')
    @require_not_blank(name='stack')
    @docparams({**doc_scale_options, **doc_stack})
    def _cmd_scale(
        self,
        replicas: str = '',
        stack: str = '',
        detach: bool = False,
        options: str = '',
    ) -> None:
        """
        Scale services (docker service scale).

        Use --replicas "svc1=3,svc2=5".
        """
        if not replicas:
            SugarLogs.raise_error(
                'Services must be provided in format service=replicas[,..]',
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        opts = self._get_list_args(options)
        if detach:
            opts.append('--detach')

        _args_extra = {'all': False, 'stack': stack}

        # alias
        fn = self._get_services_names

        pairs: list[str] = []

        for pair in replicas.split(','):
            if not pair:
                continue

            service, replica = pair.split('=')
            pairs.append(f'{fn(services=service, **_args_extra)[0]}={replica}')
        self._call_command(
            'scale', services=pairs, options_args=self._get_list_args(options)
        )

    @docparams({**doc_common_services, **doc_update_options})
    def _cmd_update(
        self,
        services: str = '',
        all: bool = False,
        detach: bool = False,
        quiet: bool = False,
        image: str = '',
        replicas: str = '',
        force: bool = False,
        rollback: bool = False,
        env_add: str = '',
        label_add: str = '',
        options: str = '',
    ) -> None:
        """Update services (docker service update)."""
        names = self._get_services_names(services=services, all=all)
        opts = self._get_list_args(options)

        if detach:
            opts.append('--detach')
        if quiet:
            opts.append('--quiet')
        if force:
            opts.append('--force')
        if rollback:
            opts.append('--rollback')
        if image:
            opts.extend(['--image', image])
        if replicas:
            opts.extend(['--replicas', replicas])

        if env_add:
            for pair in env_add.split(','):
                if pair.strip():
                    opts.extend(['--env-add', pair.strip()])
        if label_add:
            for pair in label_add.split(','):
                if pair.strip():
                    opts.extend(['--label-add', pair.strip()])

        self._call_command('update', services=names, options_args=opts)


class SugarSwarmStack(SugarSwarmBase):
    """
    SugarSwarmStack provides the docker stack commands.

    Commands: config, deploy, ls, ps, rm, services.
    """

    def _load_backend_args(self) -> None:
        self.backend_args = ['stack']

    @docparams(
        {
            **doc_stack,
            **doc_options,
        }
    )
    def _cmd_deploy(
        self,
        /,
        stack: str = '',
        options: str = '',
    ) -> None:
        """Deploy a new stack from a compose file.

        This command deploys a stack using the compose file specified
        either directly or from the profile configuration.
        """
        # Validate stack name
        if not stack:
            SugarLogs.raise_error(
                MSG_ERROR_STACK_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        # Parse options
        options_args = self._get_list_args(options)

        # Use the helper method instead of direct call
        self._call_stack_command(
            stack_name=stack,
            options_args=options_args,
            backend_args=['stack', 'deploy'],
            include_compose_files=True,
        )

    @docparams(
        {
            **doc_stack_plus_options,
            'quiet': 'Only display IDs',
        }
    )
    def _cmd_ls(
        self,
        /,
        stack: str = '',
        quiet: bool = False,
        options: str = '',
    ) -> None:
        """List the tasks in the stack."""
        if not stack:
            SugarLogs.raise_error(
                MSG_ERROR_STACK_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        backend_args = ['stack', 'ps']
        if quiet:
            backend_args = ['stack', 'ps', '--quiet']

        options_args = self._get_list_args(options)
        self._call_stack_command(
            stack_name=stack,
            options_args=options_args,
            backend_args=backend_args,
        )

    @docparams(
        {
            **doc_stack_plus_options,
            'quiet': 'Only display IDs',
        }
    )
    def _cmd_ps(
        self,
        /,
        stack: str = '',
        quiet: bool = False,
        options: str = '',
    ) -> None:
        """List the tasks in the stack."""
        if not stack:
            SugarLogs.raise_error(
                MSG_ERROR_STACK_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )

        backend_args = ['stack', 'ps']
        if quiet:
            backend_args = ['stack', 'ps', '--quiet']

        options_args = self._get_list_args(options)
        self._call_stack_command(
            stack_name=stack,
            options_args=options_args,
            backend_args=backend_args,
        )

    @docparams(doc_stack_plus_options)
    def _cmd_rm(
        self,
        /,
        stack: str = '',
        options: str = '',
    ) -> None:
        """Remove the stack from the swarm."""
        self._call_stack_command(
            stack_name=stack,
            options_args=self._get_list_args(options),
            backend_args=['stack', 'rm'],
        )


class SugarSwarmNode(SugarSwarmBase):
    """
    SugarSwarmNode provides the docker node commands.

    Commands: demote, inspect, ls, promote, ps, current node, rm, update.
    """

    def _load_backend_args(self) -> None:
        self.backend_args = ['node']

    @docparams(doc_common_nodes)
    def _cmd_demote(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """Demote one or more nodes from manager in the swarm."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command(
            'demote', nodes=node_names, options_args=options_args
        )

    @docparams(doc_common_nodes)
    def _cmd_inspect(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """Display detailed information on one or more nodes."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command(
            'inspect', nodes=node_names, options_args=options_args
        )

    @docparams(doc_common_no_services)
    def _cmd_ls(
        self,
        options: str = '',
    ) -> None:
        """List nodes in the swarm."""
        options_args = self._get_list_args(options)
        self._call_command('ls', options_args=options_args)

    @docparams(doc_common_nodes)
    def _cmd_promote(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """Promote one or more nodes to manager in the swarm."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command(
            'promote', nodes=node_names, options_args=options_args
        )

    @docparams(doc_common_nodes)
    def _cmd_ps(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """List tasks running on one or more nodes."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command('ps', nodes=node_names, options_args=options_args)

    @docparams(doc_common_nodes)
    def _cmd_rm(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """Remove one or more nodes from the swarm."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command('rm', nodes=node_names, options_args=options_args)

    @docparams(doc_common_nodes)
    def _cmd_update(
        self,
        nodes: str = '',
        options: str = '',
    ) -> None:
        """Update a node."""
        node_names = [node for node in nodes.split(',') if node]
        if not node_names:
            SugarLogs.raise_error(
                MSG_ERROR_NODES_NAME,
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        options_args = self._get_list_args(options)
        self._call_command(
            'update', nodes=node_names, options_args=options_args
        )



"""Functions about console."""

from __future__ import annotations

import os


def get_terminal_size() -> tuple[int, int]:
    """Return the width and height of the terminal using os module."""
    try:
        size = os.get_terminal_size()
        width = size.columns
        height = size.lines
    except OSError:
        # Default values if the terminal size cannot be determined.
        width = 80
        height = 24

    return width, height



"""Details screen for displaying container and service information."""

from typing import Any, TypeVar

from textual.app import ComposeResult
from textual.containers import Container, Grid, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Static,
)

# Define type variable for Screen
T = TypeVar('T')


class DetailsScreen(Screen[T]):
    """Screen to display service details."""

    BINDINGS = [
        ('escape', 'app.pop_screen', 'Back'),
        ('r', 'refresh', 'Refresh'),
        ('l', "app.push_screen('logs')", 'Logs'),
    ]

    def compose(self) -> ComposeResult:
        """Compose the details screen."""
        yield Header()

        yield Container(
            Vertical(
                Static('SERVICE DETAILS', classes='title'),
                Horizontal(
                    Static(
                        'Container and configuration information',
                        classes='subtitle',
                    ),
                    Static(id='service-name', classes='subtitle-highlight'),
                    id='screen-header',
                ),
                Grid(
                    # Configuration panel
                    Vertical(
                        Static('CONFIGURATION', classes='section-title'),
                        self._create_config_panel(),
                        classes='detail-panel',
                    ),
                    # Status panel
                    Vertical(
                        Static('STATISTICS', classes='section-title'),
                        self._create_stats_panel(),
                        classes='detail-panel',
                    ),
                    # Volumes panel
                    Vertical(
                        Static('VOLUMES', classes='section-title'),
                        self._create_volumes_panel(),
                        classes='detail-panel',
                    ),
                    # Network panel
                    Vertical(
                        Static('NETWORKS', classes='section-title'),
                        self._create_network_panel(),
                        classes='detail-panel',
                    ),
                    id='details-grid',
                ),
                Horizontal(
                    Button('View Logs', id='logs-btn', variant='primary'),
                    Button('Restart', id='restart-btn', variant='warning'),
                    Button('Stop', id='stop-btn', variant='error'),
                    Button('Back', id='back-btn', variant='default'),
                    id='details-actions',
                ),
                id='screen-content',
            ),
            id='screen-container',
        )

        yield Footer()

    def _create_config_panel(self) -> Vertical:
        """Create the configuration panel."""
        return Vertical(
            Horizontal(
                Label('Image:', classes='detail-label'),
                Label('nginx:latest', classes='detail-value', id='image-name'),
            ),
            Horizontal(
                Label('Container ID:', classes='detail-label'),
                Label(
                    'abc123def456', classes='detail-value', id='container-id'
                ),
            ),
            Horizontal(
                Label('Created:', classes='detail-label'),
                Label(
                    '2024-03-07 08:42:15',
                    classes='detail-value',
                    id='created-at',
                ),
            ),
            Horizontal(
                Label('Status:', classes='detail-label'),
                Label('● Running', classes='detail-value', id='status'),
            ),
            Horizontal(
                Label('Ports:', classes='detail-label'),
                Label('8080:80, 443:443', classes='detail-value', id='ports'),
            ),
            Horizontal(
                Label('Env Vars:', classes='detail-label'),
                Label(
                    'NGINX_HOST=example.com, PORT=80',
                    classes='detail-value',
                    id='env-vars',
                ),
            ),
            id='config-details',
        )

    def _create_stats_panel(self) -> Vertical:
        """Create the stats panel."""
        return Vertical(
            Horizontal(
                Label('CPU Usage:', classes='detail-label'),
                Label('2.5%', classes='detail-value', id='cpu-usage'),
            ),
            Horizontal(
                Label('Memory:', classes='detail-label'),
                Label(
                    '128MB / 512MB', classes='detail-value', id='memory-usage'
                ),
            ),
            Horizontal(
                Label('Network In:', classes='detail-label'),
                Label('1.2MB/s', classes='detail-value', id='network-in'),
            ),
            Horizontal(
                Label('Network Out:', classes='detail-label'),
                Label('0.8MB/s', classes='detail-value', id='network-out'),
            ),
            Horizontal(
                Label('Uptime:', classes='detail-label'),
                Label('2d 3h 45m', classes='detail-value', id='uptime'),
            ),
            Horizontal(
                Label('Restarts:', classes='detail-label'),
                Label('0', classes='detail-value', id='restarts'),
            ),
            id='stats-details',
        )

    def _create_volumes_panel(self) -> DataTable[Any]:
        """Create the volumes panel."""
        volumes_table: DataTable[Any] = DataTable(id='volumes-table')
        volumes_table.cursor_type = 'row'
        volumes_table.add_columns('Source', 'Destination', 'Mode')

        # Add mock data
        volumes_table.add_rows(
            [
                ('/data', '/var/www/html', 'rw'),
                ('nginx_logs', '/var/log/nginx', 'rw'),
                ('nginx_conf', '/etc/nginx/conf.d', 'ro'),
            ]
        )

        return volumes_table

    def _create_network_panel(self) -> DataTable[Any]:
        """Create the network panel."""
        network_table: DataTable[Any] = DataTable(id='network-table')
        network_table.cursor_type = 'row'
        network_table.add_columns('Network', 'IP Address', 'Gateway')

        # Add mock data
        network_table.add_rows(
            [
                ('bridge', '172.17.0.2', '172.17.0.1'),
                ('host', '127.0.0.1', '-'),
            ]
        )

        return network_table

    def on_mount(self) -> None:
        """Handle mounting of the screen."""
        self.query_one('#service-name', Static).update('nginx')
        self.notify('Details screen loaded')

    def action_refresh(self) -> None:
        """Refresh the details data."""
        self.notify('Refreshing service details...')

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        button_id = event.button.id

        if button_id == 'logs-btn':
            self.app.push_screen('logs')
        elif button_id == 'restart-btn':
            self.notify('Restarting service...', title='Restarting')
        elif button_id == 'stop-btn':
            self.notify('Stopping service...', title='Stopping')
        elif button_id == 'back-btn':
            self.app.pop_screen()



"""Main application module for the Sugar Terminal User Interface."""

import json

from pathlib import Path
from typing import Any, Dict, TypeVar

from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container, Grid, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import (
    Button,
    DataTable,
    Footer,
    Header,
    Label,
    Rule,
    Static,
)

from sugar.logs import SugarLogs

# Define type variables for App and Screen
A = TypeVar('A', bound=object)
T = TypeVar('T', bound=object)


class SugarTUI(App[A]):
    """Sugar Terminal User Interface application."""

    TITLE = 'Sugar TUI — Container Management Simplified'

    CSS_PATH = Path(__file__).parent / 'styles/styles.css'

    DATA_PATH = Path(__file__).parent / 'data/app.json'

    BINDINGS = [
        Binding('q', 'quit', 'Quit'),
        Binding('p', "push_screen('profiles')", 'Profiles'),
        Binding('s', "push_screen('services')", 'Services'),
        Binding('l', 'logs', 'Logs'),
        Binding('d', 'details', 'Details'),
        Binding('r', 'refresh', 'Refresh'),
        Binding('escape', 'back', 'Back'),
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.mock_data = self.load_mock_data()

    def load_mock_data(self) -> Dict[str, Any]:
        """Load mock data from JSON file."""
        try:
            with open(self.DATA_PATH, 'r') as f:
                data: Dict[str, Any] = json.load(f)
                return data
        except (FileNotFoundError, json.JSONDecodeError) as e:
            SugarLogs.print_warning(f'Error loading mock data: {e}')
            return {'profiles': [], 'services': [], 'system_metrics': {}}

    def compose(self) -> ComposeResult:
        """Compose the UI layout."""
        yield Header()

        yield Container(
            Vertical(
                Static('SUGAR TERMINAL USER INTERFACE', classes='title'),
                Static('Container Management Dashboard', classes='subtitle'),
                Grid(
                    Vertical(
                        Static('ACTIVE PROFILES', classes='dashboard-header'),
                        self._create_profiles_panel(),
                        classes='dashboard-panel profiles-panel',
                    ),
                    Vertical(
                        Static('RUNNING SERVICES', classes='dashboard-header'),
                        self._create_services_panel(),
                        classes='dashboard-panel services-panel',
                    ),
                    Vertical(
                        Static('SYSTEM METRICS', classes='dashboard-header'),
                        self._create_status_panel(),
                        classes='dashboard-panel metrics-panel',
                    ),
                    Vertical(
                        Static('QUICK ACTIONS', classes='dashboard-header'),
                        self._create_actions_panel(),
                        classes='dashboard-panel actions-panel',
                    ),
                    id='dashboard-grid',
                ),
                id='main-content',
            ),
            id='app-container',
        )

        yield Footer()

    def _create_profiles_panel(self) -> DataTable[Any]:
        """Create the profiles panel for the dashboard."""
        profiles_table: DataTable[Any] = DataTable()
        profiles_table.cursor_type = 'row'
        profiles_table.add_columns('Profile', 'Services', 'Status')

        if self.mock_data.get('profiles'):
            for profile in self.mock_data['profiles']:
                profiles_table.add_row(
                    profile['profile'], profile['services'], profile['status']
                )

        return profiles_table

    def _create_services_panel(self) -> DataTable[Any]:
        """Create the services panel for the dashboard."""
        services_table: DataTable[Any] = DataTable()
        services_table.cursor_type = 'row'
        services_table.add_columns(
            'Service', 'Status', 'Ports', 'CPU', 'Memory'
        )

        if self.mock_data.get('services'):
            for service in self.mock_data['services']:
                services_table.add_row(
                    service['service'],
                    service['status'],
                    service['ports'],
                    service['cpu'],
                    service['memory'],
                )

        return services_table

    def _create_status_panel(self) -> Vertical:
        """Create the system status panel for the dashboard."""
        metrics = self.mock_data.get('system_metrics', {})

        return Vertical(
            Horizontal(
                Label('Active Profiles:', classes='data-label'),
                Label(
                    metrics.get('active_profiles', 'N/A'), classes='data-value'
                ),
            ),
            Horizontal(
                Label('Running Containers:', classes='data-label'),
                Label(
                    metrics.get('running_containers', 'N/A'),
                    classes='data-value',
                ),
            ),
            Rule(),
            Horizontal(
                Label('CPU Usage:', classes='data-label'),
                Label(
                    metrics.get('cpu_usage', 'N/A'),
                    classes='data-value-highlight',
                ),
            ),
            Horizontal(
                Label('Memory Usage:', classes='data-label'),
                Label(
                    metrics.get('memory_usage', 'N/A'),
                    classes='data-value-highlight',
                ),
            ),
            Horizontal(
                Label('Disk Usage:', classes='data-label'),
                Label(metrics.get('disk_usage', 'N/A'), classes='data-value'),
            ),
            Rule(),
            Horizontal(
                Label('Network In:', classes='data-label'),
                Label(
                    metrics.get('network_in', 'N/A'), classes='data-value-blue'
                ),
            ),
            Horizontal(
                Label('Network Out:', classes='data-label'),
                Label(
                    metrics.get('network_out', 'N/A'),
                    classes='data-value-purple',
                ),
            ),
            Rule(),
            Horizontal(
                Label('Uptime:', classes='data-label'),
                Label(metrics.get('uptime', 'N/A'), classes='data-value'),
            ),
            classes='dashboard-data',
        )

    def _create_actions_panel(self) -> Vertical:
        """Create the quick actions panel for the dashboard."""
        return Vertical(
            Button(
                'Start All Services', id='start-all-btn', variant='success'
            ),
            Button('Stop All Services', id='stop-all-btn', variant='error'),
            Button(
                'Restart Services', id='restart-all-btn', variant='warning'
            ),
            Button('View Logs', id='view-logs-btn', variant='primary'),
            Button('Service Details', id='details-btn', variant='default'),
            Button('Run Health Checks', id='health-btn', variant='success'),
            classes='action-buttons',
        )

    def on_mount(self) -> None:
        """Event handler called when app is mounted."""
        try:
            from sugar.tui.screens.details import DetailsScreen
            from sugar.tui.screens.logs import LogsScreen
            from sugar.tui.screens.profiles import ProfileScreen
            from sugar.tui.screens.services import ServiceScreen

            self.install_screen(ProfileScreen(), name='profiles')
            self.install_screen(ServiceScreen(), name='services')
            self.install_screen(LogsScreen(), name='logs')
            self.install_screen(DetailsScreen(), name='details')

            SugarLogs.print_info('Screens installed successfully')
        except ImportError as e:
            SugarLogs.print_warning(f'Error loading screens: {e}')
            self._create_basic_screens()

    def action_refresh(self) -> None:
        """Refresh the dashboard data."""
        self.notify('Refreshing dashboard data...')
        self.mock_data = self.load_mock_data()
        self.app.refresh()

    def action_logs(self) -> None:
        """View logs."""
        self.notify('Viewing logs...')
        try:
            self.push_screen('logs')
        except Exception:
            self.notify('Logs screen not available', severity='error')

    def action_details(self) -> None:
        """View details."""
        self.notify('Viewing details...')
        try:
            self.push_screen('details')
        except Exception:
            self.notify('Details screen not available', severity='error')

    async def action_back(self) -> None:
        """Go back to previous screen."""
        try:
            self.pop_screen()
        except Exception as e:
            pass
            self.log.debug(f'Could not pop screen: {e}')

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses on the dashboard."""
        button = event.button
        button_text = button.label if hasattr(button, 'label') else ''
        button_text_str = str(button_text)

        if 'Start' in button_text_str:
            self.notify('Starting all services...', title='Starting')
        elif 'Stop' in button_text_str:
            self.notify('Stopping all services...', title='Stopping')
        elif 'Restart' in button_text_str:
            self.notify('Restarting all services...', title='Restarting')
        elif 'Logs' in button_text_str:
            self.action_logs()
        elif 'Details' in button_text_str:
            self.action_details()
        elif 'Health' in button_text_str:
            self.notify('Running health checks...', title='Health Check')

    def _create_basic_screens(self) -> None:
        """Create basic screen files if they don't exist."""
        from textual.widgets import Static

        class BasicScreen(Screen[T]):
            def compose(self) -> ComposeResult:
                yield Static('Coming soon...', classes='title')

        self.install_screen(BasicScreen(), name='profiles')
        self.install_screen(BasicScreen(), name='services')
        self.install_screen(BasicScreen(), name='logs')
        self.install_screen(BasicScreen(), name='details')

        SugarLogs.print_info('Created and installed basic screens')


def run() -> None:
    """Run the Sugar TUI application."""
    SugarLogs.print_info('Inside run() function')
    app: SugarTUI[object] = SugarTUI()
    app.run()


if __name__ == '__main__':
    run()



"""Logs classes and function for sugar system."""

from __future__ import annotations

import os

from enum import Enum

from colorama import Fore


class SugarError(Enum):
    """SugarError group all error types handled by the system."""

    SH_ERROR_RETURN_CODE = 1
    SH_KEYBOARD_INTERRUPT = 2
    SUGAR_CONFIG_FILE_NOT_FOUND = 3
    SUGAR_COMPOSE_APP_NOT_SUPPORTED = 4
    SUGAR_COMPOSE_APP_NOT_FOUNDED = 5
    SUGAR_INVALID_PARAMETER = 6
    SUGAR_MISSING_PARAMETER = 7
    SUGAR_INVALID_CONFIGURATION = 8
    SUGAR_ACTION_NOT_IMPLEMENTED = 9
    SUGAR_NO_SERVICES_RUNNING = 10
    CONFIG_VALIDATION_ERROR = 11
    YAML_PARSING_ERROR = 12
    JSON_SCHEMA_DECODING_ERROR = 13
    CONFIG_VALIDATION_UNEXPECTED_ERROR = 14
    SUGAR_COMMAND_NOT_FOUND = 15
    SUGAR_COMMAND_ERROR = 15
    SUGAR_SWARM_NOT_INITIALIZED = 16
    SUGAR_SWARM_STACK_NAME_NEEDED = 17
    SUGAR_SWARM_STACK_ALREADY_EXISTS = 18
    SUGAR_SWARM_SERVICE_NOT_FOUND = 19


class SugarLogs:
    """SugarLogs is responsible for handling system messages."""

    @staticmethod
    def raise_error(
        message: str,
        message_type: SugarError = SugarError.SH_ERROR_RETURN_CODE,
    ) -> None:
        """Print error message and exit with given error code."""
        print(Fore.RED, f'[EE] {message}', Fore.RESET)
        os._exit(message_type.value)

    @staticmethod
    def print_info(message: str) -> None:
        """Print info message."""
        print(Fore.BLUE, message, Fore.RESET)

    @staticmethod
    def print_warning(message: str) -> None:
        """Print warning message."""
        print(Fore.YELLOW, message, Fore.RESET)



"""Definition of the CLI structure."""

from __future__ import annotations

import os
import sys

from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type, Union, cast

import click
import typer

from typer import Option

from sugar import __version__
from sugar.core import extensions
from sugar.docs import MetaDocs, MetaDocsParams
from sugar.logs import SugarLogs

EXIT_CODE_CONSTANT = 2

# "count" means the number of parameters expected for each flag
CLI_ROOT_FLAGS_VALUES_COUNT = {
    '--dry-run': 0,
    '--file': 1,
    '--profile': 1,
    '--help': 0,  # not necessary to store this value
    '--verbose': 0,
    '--version': 0,  # not necessary to store this value
}

flags_state: dict[str, bool] = {
    'verbose': False,
}

flags_profile: dict[str, str] = {
    'profile': '',
}

flags_dry_run: dict[str, bool] = {
    'dry_run': False,
}

opt_state: dict[str, list[Any]] = {
    'options': [],
    'cmd': [],
}

sugar_exts = {
    ext_name: ext_class() for ext_name, ext_class in extensions.items()
}

typer_profiles: dict[str, typer.Typer] = {}

app = typer.Typer(
    name='sugar',
    help=(
        'Sugar is a tool that help you to organize'
        "and simplify your containers' stack."
    ),
    epilog=(
        'If you have any problem, open an issue at: '
        'https://github.com/osl-incubator/sugar'
    ),
    short_help="Sugar is a tool that help you \
        to organize containers' stack",
    no_args_is_help=True,
)


def _check_sugar_file(file_path: str = '.sugar.yaml') -> bool:
    return Path(file_path).exists()


def version_callback(value: bool) -> None:
    """Print the Sugar version and exit if flag is set."""
    if value:
        SugarLogs.print_info(f'Sugar version: {__version__}')
        raise typer.Exit()


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    file: str = Option(
        '',
        '--file',
        help='Set the sugar config file.',
    ),
    profile: str = Option(
        '',
        '--profile',
        help='Set the profile of services for running the sugar command.',
    ),
    version: bool = Option(
        False,
        '--version',
        '-v',
        is_flag=True,
        callback=version_callback,
        is_eager=True,
        help='Show the version of sugar.',
    ),
    verbose: bool = Option(
        False,
        '--verbose',
        is_flag=True,
        is_eager=True,
        help='Show the command executed.',
    ),
    dry_run: bool = Option(
        False,
        '--dry-run',
        is_flag=True,
        is_eager=True,
        help="Don't actually execute the command.",
    ),
) -> None:
    """
    Process commands for specific flags.

    Otherwise, show the help menu.
    """
    ctx.ensure_object(dict)

    if verbose:
        # global
        flags_state['verbose'] = True

    if profile:
        # global
        flags_profile['profile'] = profile

    if dry_run:
        # global
        flags_dry_run['dry_run'] = True

    if ctx.invoked_subcommand is None:
        raise typer.Exit()


def map_type_from_string(type_name: str) -> Type[Union[str, int, float, bool]]:
    """
    Return a type object mapped from the type name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    type
        The corresponding Python type.
    """
    type_mapping = {
        'str': str,
        'string': str,
        'int': int,
        'integer': int,
        'float': float,
        'bool': bool,
        'boolean': bool,
    }
    return type_mapping.get(type_name, str)


def normalize_string_type(type_name: str) -> str:
    """
    Normalize the user type definition to the correct name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    str
        The corresponding makim type name.
    """
    type_mapping = {
        'str': 'str',
        'string': 'str',
        'int': 'int',
        'integer': 'int',
        'float': 'float',
        'bool': 'bool',
        'boolean': 'bool',
        # Add more mappings as needed
    }
    return type_mapping.get(type_name, 'str')


def get_default_value(
    arg_type: str, value: Any
) -> Optional[Union[str, int, float, bool]]:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'bool':
        return False if value is None else bool(value)
    elif arg_type == 'int':
        return int(value) if value is not None else None
    elif arg_type == 'float':
        return float(value) if value is not None else None
    elif arg_type == 'str':
        return str(value) if value is not None else None
    return None


def get_default_value_str(arg_type: str, value: Any) -> str:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'str':
        return f'"{value}"'

    if arg_type == 'bool':
        return 'False'

    return f'{value or 0}'


def create_args_string(args: dict[str, dict[str, str]]) -> str:
    """Return a string for arguments for a function for typer."""
    args_rendered = []

    arg_template_flag = (
        '{arg_name}: Optional[{arg_type}] = typer.Option('
        '{default_value}, '
        '"--{name_flag}", '
        'help="{help_text}"'
        ')'
    )

    arg_template_positional = (
        '{arg_name}: Optional[{arg_type}] = typer.Argument('
        '{default_value}, '
        'help="{help_text}"'
        ')'
    )

    for name, spec in args.items():
        name_clean = name.replace('-', '_')
        arg_type = normalize_string_type(spec.get('type', 'str'))
        help_text = spec.get('help', '')
        default_value = 'None'
        is_positional_only = (
            spec.get('positional_only', 'true').lower() == 'true'
        )

        if not spec.get('required', False) and not spec.get(
            'interactive', False
        ):
            default_value = spec.get('default', '')
            default_value = get_default_value_str(arg_type, default_value)

        selected_template = (
            arg_template_positional
            if is_positional_only
            else arg_template_flag
        )

        arg_str = selected_template.format(
            **{
                'arg_name': name_clean,
                'arg_type': arg_type,
                'default_value': default_value,
                'name_flag': name,
                'help_text': help_text.replace('\n', '\\n'),
            }
        )

        args_rendered.append(arg_str)

    return ', '.join(args_rendered)


def apply_click_options(
    command_function: Callable[..., Any], options: dict[str, Any]
) -> Callable[..., Any]:
    """
    Apply Click options to a Typer command function.

    Parameters
    ----------
    command_function : callable
        The Typer command function to which options will be applied.
    options : dict
        A dictionary of options to apply.

    Returns
    -------
    callable
        The command function with options applied.
    """
    for opt_name, opt_details in options.items():
        opt_args: dict[
            str, Optional[Union[str, int, float, bool, Type[Any]]]
        ] = {}

        opt_data = cast(Dict[str, str], opt_details)
        opt_type_str = normalize_string_type(opt_data.get('type', 'str'))
        opt_default = get_default_value(opt_type_str, opt_data.get('default'))

        if opt_type_str == 'bool':
            opt_args.update({'is_flag': True})

        opt_args.update(
            {
                'default': None
                if opt_data.get('interactive', False)
                else opt_default,
                'type': map_type_from_string(opt_type_str),
                'help': opt_data.get('help', ''),
                'show_default': True,
            }
        )

        click_option = click.option(
            f'--{opt_name}',
            **opt_args,  # type: ignore[arg-type]
        )
        command_function = click_option(command_function)

    return command_function


def create_dynamic_command(
    ext_name: str,
    typer_profile: typer.Typer,
    meta: MetaDocs,
) -> None:
    """
    Dynamically create a Typer command with the specified options.

    Parameters
    ----------
    typer_profile : typer.Typer
    name : str
        The command name.
    meta : dict
        the action/command metadata
    """
    name = cast(str, meta.get('name', ''))
    args = cast(Dict[str, Dict[str, str]], meta.get('parameters', {}))
    fn_help = cast(str, meta.get('title', ''))

    args_str = create_args_string(args)
    args_param_list: list[str] = []
    for arg, _spec in args.items():
        arg_clean = arg.replace('-', '_')
        args_param_list.append(f'{arg}={arg_clean}')
    args_param_str = ','.join(args_param_list)

    # IMPORTANT: allow extra args so we can capture tokens after `--`
    decorator = typer_profile.command(
        name,
        help=fn_help,
        context_settings={
            'allow_extra_args': True,
            'ignore_unknown_options': True,
        },
    )

    # we need ctx to access ctx.args (the tail after `--`)
    needs_options_glue = (
        'options' in args
    )  # only glue tail if command has `options`

    # build function signature
    params_sig = 'ctx: typer.Context'
    if args_str:
        params_sig += f', {args_str}'

    # build function body
    if needs_options_glue:
        glue = (
            '    # If the user typed a literal `--`, '
            'treat the tail as backend options\n'
            '    # Click removes the `--` token, so we detect it via sys.argv '
            'and read ctx.args\n'
            "    extra_tail = list(ctx.args) if '--' in sys.argv else []\n"
            '    if extra_tail:\n'
            '        # merge tail into the `options` string (works with your '
            'existing _get_list_args)\n'
            "        options = (options or '').strip()\n"
            "        tail_str = ' '.join(extra_tail)\n"
            "        options = f'{options} {tail_str}'.strip() if options "
            'else tail_str\n'
        )
        call_line = f'    sugar._cmd_{name}({args_param_str})\n'
    else:
        glue = (
            '    # Command has no `options` parameter; ignore tail to avoid '
            'swallowing typos\n'
            '    _ = ctx.args  # still available if you later want to use it\n'
        )
        call_line = f'    sugar._cmd_{name}({args_param_str})\n'

    function_code = (
        f'def dynamic_command({params_sig}):\n'
        f"    sugar = sugar_exts.get('{ext_name}')\n"
        f'{glue}'
        f'{call_line}'
    )

    local_vars: dict[str, Any] = {}
    exec(function_code, globals(), local_vars)
    decorator(local_vars['dynamic_command'])


# function to create a callback for each Typer profile
def subcommand_callback() -> Callable[..., None]:
    """Create a callback function for a subcommand Typer instance."""

    def callback(ctx: typer.Context) -> None:
        """Show help when subcommand is invoked without an operation."""
        if ctx.invoked_subcommand is None:
            ctx.obj = ctx.obj or {}
            typer.echo(ctx.command.get_help(ctx))
            raise typer.Exit()

    return callback


def extract_root_config(
    cli_list: list[str] = sys.argv,
) -> dict[str, str | bool]:
    """Extract the root configuration from the CLI."""
    params = cli_list[1:]

    # default values
    sugar_file = '.sugar.yaml'
    profile = ''
    dry_run = False
    verbose = False

    try:
        idx = 0
        while idx < len(params):
            arg = params[idx]
            if arg not in CLI_ROOT_FLAGS_VALUES_COUNT:
                break

            if arg == '--file':
                try:
                    sugar_file = params[idx + 1]
                except IndexError:
                    pass
            elif arg == '--profile':
                try:
                    profile = params[idx + 1]
                except IndexError:
                    pass
            elif arg == '--dry-run':
                dry_run = True
            elif arg == '--verbose':
                verbose = True

            idx += 1 + CLI_ROOT_FLAGS_VALUES_COUNT[arg]
    except Exception:
        red_text = typer.style(
            'The sugar config file was not correctly detected. '
            'Using the default .sugar.yaml.',
            fg=typer.colors.RED,
            bold=True,
        )
        typer.echo(red_text, err=True, color=True)

    return {
        'file': sugar_file,
        'profile': profile,
        'dry_run': dry_run,
        'verbose': verbose,
    }


def _get_command_from_cli() -> str:
    """
    Get the profile and task from CLI.

    This function is based on `CLI_ROOT_FLAGS_VALUES_COUNT`.
    """
    params = sys.argv[1:]
    command = ''

    try:
        idx = 0
        while idx < len(params):
            arg = params[idx]
            if arg not in CLI_ROOT_FLAGS_VALUES_COUNT:
                command = f'flag `{arg}`' if arg.startswith('--') else arg
                break

            idx += 1 + CLI_ROOT_FLAGS_VALUES_COUNT[arg]
    except Exception as e:
        print(e)

    return command


def _is_help_requested() -> bool:
    """Check if help flag is present in arguments."""
    return any(arg in ['--help', '-h'] for arg in sys.argv[1:])


def _handle_config_file(help_requested: bool, config_file_path: str) -> None:
    """Handle config file validation and errors."""
    if not help_requested and not _check_sugar_file(config_file_path):
        typer.secho(
            f"Error: Sugar config file '{config_file_path}' "
            'not found in current directory.',
            fg='red',
            err=True,
        )
        raise typer.Exit(1)


def _show_warning_config_file_not_found(config_file_path: str) -> None:
    if not _check_sugar_file(config_file_path):
        typer.secho(
            f"Warning: Sugar config file '{config_file_path}' "
            'not found. Some commands may not be fully documented.',
            fg='yellow',
            err=True,
        )


def _handle_autocomplete(
    help_requested: bool, config_file_path: str
) -> tuple[dict[str, str | bool], str]:
    """Handle autocomplete when config file is not found."""
    cli_completion_words = [
        w for w in os.getenv('COMP_WORDS', '').split('\n') if w
    ]

    root_config = extract_root_config()

    if (
        not help_requested
        and not _check_sugar_file(config_file_path)
        and cli_completion_words
    ):
        # autocomplete call
        root_config = extract_root_config(cli_completion_words)
        config_file_path = cast(str, root_config.get('file', '.sugar.yaml'))

    return root_config, config_file_path


def _process_extensions(
    help_requested: bool,
    config_file_path: str,
    root_config: dict[str, str | bool],
) -> dict[str, list[MetaDocs]]:
    """Process extensions and extract commands."""
    for sugar_ext in sugar_exts.values():
        if not help_requested:
            sugar_ext.load(
                file=config_file_path,
                profile=cast(str, root_config.get('profile', '')),
                dry_run=cast(bool, root_config.get('dry_run', False)),
                verbose=cast(bool, root_config.get('verbose', False)),
            )

    commands: dict[str, list[MetaDocs]] = {}

    for ext_name, ext_class in extensions.items():
        ext_obj = ext_class()
        commands[ext_name] = []

        for action in ext_obj.actions:
            fn_name = f'_cmd_{action}'
            fn = getattr(ext_obj, fn_name)
            title = fn._meta_docs.get('title', '')

            commands[ext_name].append(
                cast(
                    MetaDocs,
                    {
                        'name': action,
                        'title': title,
                        'parameters': cast(
                            MetaDocsParams, fn._meta_docs.get('parameters', {})
                        ),
                    },
                )
            )

    return commands


def _setup_typer_app(commands: dict[str, list[MetaDocs]]) -> None:
    """Set up the Typer app with commands."""
    for ext_name, actions_meta in commands.items():
        ext_obj = extensions[ext_name]()

        if not ext_obj:
            SugarLogs.raise_error(f'Extension not found ({ext_name}).')

        typer_profile = typer.Typer(
            help=ext_obj.__doc__,
            invoke_without_command=True,
        )

        # callback to show help when subcommand is invoked without operation
        typer_profile.callback()(subcommand_callback())

        typer_profiles[ext_name] = typer_profile

        for action_meta in actions_meta:
            create_dynamic_command(ext_name, typer_profile, action_meta)

    # Add each profile to the main app
    for ext_name, typer_profile in typer_profiles.items():
        app.add_typer(
            typer_profile,
            name=ext_name,
            rich_help_panel='COMMAND',
            no_args_is_help=True,
        )


def run_app() -> None:
    """Run the typer app."""
    root_config = extract_root_config()
    help_requested = _is_help_requested()

    config_file_path = cast(str, root_config.get('file', '.sugar.yaml'))

    # Show warning if config file not found while showing help
    _show_warning_config_file_not_found(config_file_path)

    # Handle autocomplete
    root_config, config_file_path = _handle_autocomplete(
        help_requested, config_file_path
    )

    if not _check_sugar_file(config_file_path) and not help_requested:
        return

    _handle_config_file(help_requested, config_file_path)

    # Process extensions and extract commands
    commands = _process_extensions(
        help_requested, config_file_path, root_config
    )

    # Set up Typer app with commands
    _setup_typer_app(commands)

    try:
        app()
    except SystemExit as e:
        # code 2 means command not found
        if e.code == EXIT_CODE_CONSTANT:
            app(['--help'], standalone_mode=False)
            typer.Exit(code=1)
            os._exit(1)
        raise e


if __name__ == '__main__':
    run_app()



"""Log viewer screen for monitoring service and container logs."""

import random  # nosec B311 - not used for security/crypto purposes, just for demo data

from datetime import datetime
from pathlib import Path
from typing import Optional, TypeVar

from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Static

MAX_LOG_LINES = 20

# Define type variable for Screen
T = TypeVar('T', bound=object)


class LogsScreen(Screen[T]):
    """Screen to display service logs with filtering and real-time updates."""

    # Match the same CSS path
    CSS_PATH = Path(__file__).parent.parent / 'styles/screens.css'

    BINDINGS = [
        ('escape', 'app.pop_screen', 'Back'),
        ('f', 'toggle_follow', 'Follow'),
        ('c', 'clear_logs', 'Clear'),
        ('r', 'refresh', 'Refresh'),
    ]

    def __init__(self) -> None:
        super().__init__()
        self.following = False
        self.service_name = 'frontend'
        self.log_levels = ['INFO', 'WARN', 'ERROR', 'DEBUG']

    def compose(self) -> ComposeResult:
        """Compose the logs screen layout."""
        yield Header()

        yield Container(
            Vertical(
                Static('SERVICE LOGS', classes='title'),
                Horizontal(
                    Static('Real-time container logs', classes='subtitle'),
                    Static(
                        'Auto-refresh: On',
                        id='refresh-status',
                        classes='auto-refresh-on',
                    ),
                    id='screen-header',
                ),
                Horizontal(
                    Button('All', id='filter-all', variant='primary'),
                    Button('INFO', id='filter-info', variant='default'),
                    Button('WARN', id='filter-warn', variant='warning'),
                    Button('ERROR', id='filter-error', variant='error'),
                    Button('DEBUG', id='filter-debug', variant='success'),
                    id='log-level-filters',
                ),
                # Simple logs display using a Static widget instead of Log
                Static(id='logs-display', classes='logs-display'),
                Horizontal(
                    Button('Back', id='back-btn', variant='default'),
                    Button('Follow', id='follow-btn', variant='primary'),
                    Button('Clear', id='clear-btn', variant='warning'),
                    id='logs-actions',
                ),
                id='screen-content',
            ),
            id='screen-container',
        )

        yield Footer()

    def on_mount(self) -> None:
        """Handle mounting of the screen."""
        self.notify('Logs screen loaded')
        self.generate_sample_logs()

        # Set up auto-refresh timer for log updates
        self.set_interval(3, self.add_new_log_entry)

    def generate_sample_logs(self) -> None:
        """Generate sample log entries."""
        logs_display = self.query_one('#logs-display', Static)

        # Sample log entries with varied severity levels and timestamps
        sample_logs = [
            self.format_log_entry(
                '2024-03-09 10:00:01', 'INFO', 'Application starting...'
            ),
            self.format_log_entry(
                '2024-03-09 10:00:02',
                'INFO',
                'Loading configuration from /etc/config.json',
            ),
            self.format_log_entry(
                '2024-03-09 10:00:03',
                'DEBUG',
                'Configuration loaded: { "port": 3000, "env": "production" }',
            ),
            self.format_log_entry(
                '2024-03-09 10:00:05',
                'INFO',
                'Connecting to database at db:5432',
            ),
            self.format_log_entry(
                '2024-03-09 10:00:07',
                'INFO',
                'Database connection established',
            ),
            self.format_log_entry(
                '2024-03-09 10:00:10',
                'INFO',
                'Starting web server on port 3000',
            ),
            self.format_log_entry(
                '2024-03-09 10:00:15',
                'INFO',
                'Server listening at http://0.0.0.0:3000',
            ),
            self.format_log_entry(
                '2024-03-09 10:01:02',
                'INFO',
                'Received request GET /api/status',
            ),
            self.format_log_entry(
                '2024-03-09 10:01:03',
                'DEBUG',
                'Request headers: { "auth": "***" }',
            ),
            self.format_log_entry(
                '2024-03-09 10:01:04', 'INFO', 'Response sent: 200 OK (15ms)'
            ),
        ]

        # Join log entries with newlines and update the display
        logs_text = '\n'.join(sample_logs)
        logs_display.update(logs_text)

    def format_log_entry(
        self, timestamp: str, level: str, message: str
    ) -> str:
        """Format a log entry."""
        timestamp_color = 'cyan'
        level_color = {
            'INFO': 'green',
            'WARN': 'yellow',
            'ERROR': 'red',
            'DEBUG': 'blue',
        }.get(level, 'white')

        formatted_msg = f'[{timestamp_color}]{timestamp}[/]'
        formatted_msg += f' [{level_color}]{level:7}[/] {message}'
        return formatted_msg

    def add_new_log_entry(self) -> None:
        """Add a new log entry if following is enabled."""
        if not self.following:
            return

        logs_display = self.query_one('#logs-display', Static)
        # Convert the renderable to string before splitting
        current_logs = str(logs_display.render())

        # Current timestamp
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        # Generate demo log data (not for security purposes)
        level = random.choice(self.log_levels)  # nosec B311

        # Sample messages based on log level
        messages = {
            'INFO': [
                'Received request GET /api/status',
                f'Processing batch job #{random.randint(1000, 9999)}',  # nosec B311
                f'Response sent: 200 OK ({random.randint(5, 150)}ms)',  # nosec B311
            ],
            'WARN': [
                f'High CPU usage detected: {random.randint(70, 95)}%',  # nosec B311
                f'Slow query detected ({random.randint(1, 10)}s)',  # nosec B311
            ],
            'ERROR': [
                'Failed to connect to service: connection refused',
                f'Database query timeout after {random.randint(10, 60)}s',  # nosec B311
            ],
            'DEBUG': [
                f'Request parameters: {{ "id": {random.randint(1, 1000)} }}',  # nosec B311
                f'Cache hit ratio: {random.randint(60, 95)}%',  # nosec B311
            ],
        }

        # Select a random message for demo purposes
        message = random.choice(messages[level])  # nosec B311

        # Format and add the log entry
        log_entry = self.format_log_entry(timestamp, level, message)

        log_lines = current_logs.split('\n')
        log_lines.append(log_entry)
        if len(log_lines) > MAX_LOG_LINES:
            log_lines = log_lines[-MAX_LOG_LINES:]

        logs_display.update('\n'.join(log_lines))

    def action_toggle_follow(self) -> None:
        """Toggle log following."""
        self.following = not self.following
        button = self.query_one('#follow-btn', Button)
        button.variant = 'success' if self.following else 'primary'
        button.label = 'Following' if self.following else 'Follow'

        status_label = (
            'Auto-refresh: On' if self.following else 'Auto-refresh: Off'
        )
        self.query_one('#refresh-status', Static).update(status_label)
        self.query_one('#refresh-status').set_classes(
            'auto-refresh-on' if self.following else 'auto-refresh-off'
        )

        status = 'enabled' if self.following else 'disabled'
        self.notify(f'Log following {status}')

    def action_clear_logs(self) -> None:
        """Clear all logs."""
        logs_display = self.query_one('#logs-display', Static)
        logs_display.update('')
        self.notify('Logs cleared')

    def apply_filter(self, level: Optional[str] = None) -> None:
        """Apply a log level filter."""
        # Reset button variants
        for btn_id in [
            'filter-all',
            'filter-info',
            'filter-warn',
            'filter-error',
            'filter-debug',
        ]:
            btn = self.query_one(f'#{btn_id}', Button)
            btn.variant = 'default'

        # Set active filter button
        if level:
            self.query_one(
                f'#filter-{level.lower()}', Button
            ).variant = 'primary'
        else:
            self.query_one('#filter-all', Button).variant = 'primary'

        # Apply filter logic here (in a real implementation)
        filter_text = f'level: {level}' if level else 'all levels'
        self.notify(f'Filter applied: {filter_text}')

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        button_id = event.button.id

        if button_id == 'follow-btn':
            self.action_toggle_follow()
        elif button_id == 'clear-btn':
            self.action_clear_logs()
        elif button_id == 'back-btn':
            self.app.pop_screen()
        elif button_id == 'filter-all':
            self.apply_filter()
        elif button_id == 'filter-info':
            self.apply_filter('INFO')
        elif button_id == 'filter-warn':
            self.apply_filter('WARN')
        elif button_id == 'filter-error':
            self.apply_filter('ERROR')
        elif button_id == 'filter-debug':
            self.apply_filter('DEBUG')



"""SugarComposeExt Plugin class for containers."""

from __future__ import annotations

from sugar.docs import docparams
from sugar.extensions.compose import (
    SugarCompose,
    doc_common_services,
)


class SugarComposeExt(SugarCompose):
    """SugarComposeExt provides extra commands on top of docker-compose."""

    @docparams(doc_common_services)
    def _cmd_restart(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Restart services (compose stop + up)."""
        self._cmd_stop(services=services, all=all)
        self._cmd_start(services=services, all=all, options=options)

    @docparams(doc_common_services)
    def _cmd_start(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Start services (compose up)."""
        self._cmd_up(services=services, all=all, options=options)



"""Sugar class for containers."""

from __future__ import annotations

from typing import Optional, Type, cast

from sugar.extensions.base import SugarBase
from sugar.extensions.compose import SugarCompose
from sugar.extensions.compose_ext import SugarComposeExt
from sugar.extensions.podman_ext import SugarPodmanComposeExt
from sugar.extensions.swarm import (
    SugarSwarm,
    SugarSwarmNode,
    SugarSwarmService,
    SugarSwarmStack,
)
from sugar.extensions.tui import SugarTUI

try:
    from sugar.extensions.stats import SugarStats
except ImportError:
    # SugarStats is optional (extras=tui)
    SugarStats = cast(Optional[Type[SugarBase]], None)  # type: ignore


extensions: dict[str, Type[SugarBase]] = {
    'compose': SugarCompose,
    'compose-ext': SugarComposeExt,
    'podman-ext': SugarPodmanComposeExt,
    'swarm': SugarSwarm,
    'swarm:stack': SugarSwarmStack,
    'swarm:service': SugarSwarmService,
    'swarm:node': SugarSwarmNode,
    'tui': SugarTUI,
    **{'stats': SugarStats for i in range(1) if SugarStats is not None},
}



"""Sugar plugin for podman compose."""

from __future__ import annotations

import io
import os
import sys

from pathlib import Path
from typing import Any, Union

import dotenv
import sh
import yaml

from sugar.docs import docparams
from sugar.extensions.base import SugarBase
from sugar.logs import SugarError, SugarLogs
from sugar.utils import (
    camel_to_snake,
    get_absolute_path,
)

doc_profile = {
    'profile': 'Specify the profile name of the services you want to use.'
}
doc_service = {'service': 'Set the service for the container call.'}
doc_services = {
    'services': 'Set the services separated by comma for the container call.'
}
doc_all_services = {'all': 'Use all services for the command.'}
doc_options = {
    'options': ('Specify the options for the backend command. E.g.: `--  -d`.')
}
doc_cmd = {
    'cmd': (
        'Specify the COMMAND for some podman-compose command. '
        'E.g.: --cmd python -c print(1).'
    )
}

doc_common_no_services = {
    **doc_profile,
    **doc_options,
}

doc_common_service = {
    **doc_profile,
    **doc_service,
    **doc_options,
}

doc_common_service_cmd = {
    **doc_profile,
    **doc_service,
    **doc_options,
    **doc_cmd,
}

doc_common_services = {
    **doc_profile,
    **doc_services,
    **doc_all_services,
    **doc_options,
}

doc_common_services_no_options = {
    **doc_profile,
    **doc_services,
    **doc_all_services,
}


class SugarPodmanComposeExt(SugarBase):
    """(Experimental)SugarPodmanCompose provide the podman compose commands."""

    def _load_backend(self) -> None:
        """
        Initialize the backend application and its arguments.

        This method checks if the backend is supported, sets the backend app,
        and loads backend arguments.
        """
        backend_cmd = self.config.get('backend', '')
        supported_backends = ['podman', 'compose']

        if backend_cmd not in supported_backends:
            SugarLogs.raise_error(
                f'"{self.config["backend"]}" not supported yet.'
                f' Supported backends are: {", ".join(supported_backends)}.',
                SugarError.SUGAR_COMPOSE_APP_NOT_SUPPORTED,
            )

        self.backend_app = sh.Command('podman-compose')
        self._load_podman_compose_args()

    def _load_podman_compose_args(self) -> None:
        """Load arguments specific to podman-compose."""
        self._filter_service_profile()

        config_path = []

        if type(self.service_profile['config-path']) is list:
            backend_path_arg = [
                get_absolute_path(path)
                for path in self.service_profile.get('config-path', [])
            ]
        else:
            backend_path_arg = [
                get_absolute_path(self.service_profile['config-path'])
            ]

        if isinstance(backend_path_arg, str):
            # Convert relative path to absolute using project root
            if not backend_path_arg.startswith('/'):
                backend_path_arg = str(
                    Path(self.file).parent / backend_path_arg
                )
            config_path.append(backend_path_arg)
        elif isinstance(backend_path_arg, list):
            config_path.extend(backend_path_arg)
        else:
            SugarLogs.raise_error(
                'The attribute config-path` just supports the data '
                f'types `string` or `list`, {type(backend_path_arg)} '
                'received.',
                SugarError.SUGAR_INVALID_CONFIGURATION,
            )
        # Use -f instead of --file for podman compose
        for p in config_path:
            self.backend_args.extend(['-f', p])

        if self.service_profile.get('project-name'):
            self.backend_args.extend(
                ['-p', self.service_profile['project-name']]
            )

    def _call_backend_app(
        self,
        action: str,
        services: list[str] = [],
        nodes: list[str] = [],
        options_args: list[str] = [],
        cmd_args: list[str] = [],
        _out: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stdout,
        _err: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stderr,
    ) -> None:
        """
        Override _call_backend_app to handle environment variables.

        from .env file.
        instead of using the --env-file flag that's
        not supported by podman compose.
        """
        # validation
        if nodes:
            SugarLogs.raise_error(
                'The argument `nodes` is not supported yet '
                'by the podman plugin.',
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        # Execute pre-run hooks
        extension = camel_to_snake(
            self.__class__.__name__.replace('Sugar', '')
        )
        self._execute_hooks('pre-run', extension, action)

        # Load env vars from env-file if specified
        env_vars = os.environ.copy()
        env_file = self.service_profile.get('env-file')
        if env_file:
            if not env_file.startswith('/'):
                env_file = str(Path(self.file).parent / env_file)

            if Path(env_file).exists():
                try:
                    dotenv_values = {
                        k: str(v)
                        for k, v in dotenv.dotenv_values(env_file).items()
                        if v is not None
                    }
                    env_vars.update(dotenv_values)
                    if self.verbose:
                        SugarLogs.print_info(
                            f'Loaded environment variables from {env_file}'
                        )
                except ImportError:
                    SugarLogs.print_warning(
                        """python-dotenv package not found.
                        Env file variables will not be loaded."""
                    )
                except Exception as e:
                    SugarLogs.print_warning(
                        f'Failed to load env file {env_file}: {e!s}'
                    )

        sh_extras = {
            '_in': sys.stdin,
            '_out': _out,
            '_err': _err,
            '_no_err': True,
            '_env': env_vars,
            '_bg': True,
            '_bg_exc': False,
        }

        positional_parameters = [
            *self.backend_args,
            *[action],
            *options_args,
            *services,
            *cmd_args,
        ]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> {self.backend_app} {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        p = self.backend_app(
            *positional_parameters,
            **sh_extras,
        )

        try:
            p.wait()
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            pid = p.pid
            p.kill()
            SugarLogs.raise_error(
                f'Process {pid} killed.', SugarError.SH_KEYBOARD_INTERRUPT
            )

        self._execute_hooks('post-run', extension, action)

    def podman_check_services(
        self,
        all: bool,
        services: str,
    ) -> list[str]:
        """Check if difference between default and available podman services.

        Check if difference between default and available
        podman services in .sugar.yml file.
        """
        if all:
            default = self._get_services_names(services=services, all=True)
            available = self._get_services_names(services=services, all=False)
            diffence = set(default) - set(available)
            return list(diffence)
        else:
            default = self._get_services_names(services=services, all=False)
            return default

    def _get_image_filters(self, services_names: list[str]) -> list[str]:
        """Get image filters for the given services."""
        filters = []
        config_paths = self.service_profile.get('config-path', [])
        if isinstance(config_paths, str):
            config_paths = [config_paths]

        for config_path in config_paths:
            path_to_use = config_path
            if not config_path.startswith('/'):
                path_to_use = str(Path(self.file).parent / config_path)
            try:
                with open(path_to_use) as f:
                    compose_config = yaml.safe_load(f)
                    for service in services_names:
                        if service in compose_config.get('services', {}):
                            service_config = compose_config['services'][
                                service
                            ]
                            if 'image' in service_config:
                                filters.extend(
                                    [
                                        '--filter',
                                        f'reference={service_config["image"]}',
                                    ]
                                )
            except (yaml.YAMLError, FileNotFoundError) as e:
                SugarLogs.print_warning(
                    f'Error with compose file {path_to_use}: {e}'
                )
        return filters

    @docparams(doc_common_service)
    def _cmd_attach(
        self,
        service: str = '',
        options: str = '',
    ) -> None:
        """
        Attach to a service's container.

        Attach local standard input, output, and error streams to a
        running container.
        """
        if not service:
            SugarLogs.raise_error(
                'The service parameter is required.',
                SugarError.SUGAR_MISSING_PARAMETER,
            )

        # Use podman command directly instead of podman-compose
        podman_cmd = sh.Command('podman')
        options_args = self._get_list_args(options)

        # Get the container name based on the service name
        # This assumes container name follows podman-compose naming convention
        project_name = self.service_profile.get('project-name', 'sugar')
        container_name = f'{project_name}_{service}_1'

        # Build command arguments
        positional_parameters = ['attach', *options_args, container_name]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> podman {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        try:
            podman_cmd(
                *positional_parameters,
                _fg=True,  # Run in foreground to allow direct interaction
            )
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            SugarLogs.raise_error(
                'Detached from container.', SugarError.SH_KEYBOARD_INTERRUPT
            )

    @docparams(doc_common_services)
    def _cmd_build(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Build or rebuild services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'build', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_config(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Parse, resolve and render compose file in canonical format."""
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'config',
            services=[],
            options_args=options_args,
        )

    @docparams(doc_common_no_services)
    def _cmd_cp(
        self,
        options: str = '',
    ) -> None:
        """
        Copy files/folders between a container and the local filesystem.

        Use SRC_PATH and DEST_PATH format where one of them should be a
        container name.
        Example: container_name:/path/to/file /local/path
        """
        # Use podman command directly instead of podman-compose
        podman_cmd = sh.Command('podman')
        options_args = self._get_list_args(options)

        # Build command arguments
        positional_parameters = [
            'cp',
            *options_args,
        ]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> podman {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        try:
            podman_cmd(
                *positional_parameters,
                _fg=True,  # Run in foreground to allow direct interaction
            )
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            SugarLogs.raise_error(
                'Copy operation interrupted.', SugarError.SH_KEYBOARD_INTERRUPT
            )

    @docparams(doc_common_services)
    def _cmd_create(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Create services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'run', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_down(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Stop and remove containers, networks."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'down', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_events(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """
        Monitor podman events.

        Shows events from containers matching the specified services.
        Use --filter to filter output (e.g. --  --filter event=create).
        """
        if not services and not all:
            SugarLogs.raise_error(
                'Either service names or --all flag is required.',
                SugarError.SUGAR_MISSING_PARAMETER,
            )

        # Use podman command directly instead of podman-compose
        podman_cmd = sh.Command('podman')
        options_args = self._get_list_args(options)
        services_names = self._get_services_names(services=services, all=all)

        # Get project name for container name prefixing
        project_name = self.service_profile.get('project-name', 'sugar')

        # Build container filters for each service
        filters = []
        for service in services_names:
            container_name = f'{project_name}_{service}_1'
            filters.extend(['--filter', f'container={container_name}'])

        # Build command arguments
        positional_parameters = [
            'events',
            *filters,
            *options_args,
        ]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> podman {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        try:
            podman_cmd(
                *positional_parameters,
                _fg=True,  # Run in foreground for live event stream
            )
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            SugarLogs.raise_error(
                'Events monitoring interrupted.',
                SugarError.SH_KEYBOARD_INTERRUPT,
            )

    @docparams(doc_common_service_cmd)
    def _cmd_exec(
        self,
        service: str = '',
        cmd: str = '',
        options: str = '',
    ) -> None:
        """Execute a command in a running container."""
        if not service:
            SugarLogs.raise_error(
                'The service parameter is required.',
                SugarError.SUGAR_MISSING_PARAMETER,
            )

        options_args = self._get_list_args(options)
        cmd_args = self._get_list_args(cmd)
        self._call_backend_app(
            'exec',
            services=[service],
            options_args=options_args,
            cmd_args=cmd_args,
        )

    @docparams(doc_common_services)
    def _cmd_images(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """List images used by the created containers."""
        if not services and not all:
            SugarLogs.raise_error(
                'Either service names or --all flag is required.',
                SugarError.SUGAR_MISSING_PARAMETER,
            )

        podman_cmd = sh.Command('podman')
        options_args = self._get_list_args(options)
        services_names = self._get_services_names(services=services, all=all)
        filters = self._get_image_filters(services_names)

        positional_parameters = ['images', *filters, *options_args]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> podman {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        try:
            podman_cmd(*positional_parameters, _fg=True)
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarError.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            SugarLogs.raise_error(
                'Image listing interrupted.', SugarError.SH_KEYBOARD_INTERRUPT
            )

    @docparams(doc_common_services)
    def _cmd_kill(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Kill containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'kill', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_logs(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """View output from containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'logs', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_pause(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Pause services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'pause', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_port(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Print the public port for a port binding."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'port', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_ps(
        self,
        options: str = '',
    ) -> None:
        """List containers."""
        # Podman-compose ps doesn't support filtering by services
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'ps',
            services=[],
            options_args=options_args,
        )

    @docparams(doc_common_services)
    def _cmd_pull(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Pull service images."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'pull', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_push(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Push service images."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'push', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_restart(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Restart services."""
        if '-d' in options.split(' '):
            SugarLogs.raise_error(
                'The -d option is not supported for the restart command.',
                SugarError.SUGAR_INVALID_PARAMETER,
            )
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'restart', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_rm(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Remove stopped containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args('-v ' + options)
        self._call_backend_app(
            'down', services=services_names, options_args=options_args
        )

    @docparams(doc_common_service_cmd)
    def _cmd_run(
        self,
        service: str = '',
        cmd: str = '',
        options: str = '',
    ) -> None:
        """Run a one-off command on a service."""
        if not service:
            SugarLogs.raise_error(
                'The service parameter is required.',
                SugarError.SUGAR_MISSING_PARAMETER,
            )

        options_args = self._get_list_args(options)
        cmd_args = self._get_list_args(cmd)

        self._call_backend_app(
            'run',
            services=[service],
            options_args=options_args,
            cmd_args=cmd_args,
        )

    @docparams(doc_common_services)
    def _cmd_start(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Start services."""
        options_args = self._get_list_args(options)
        diff_service = self.podman_check_services(services=services, all=all)
        if diff_service != []:
            SugarLogs.print_info(
                'Running un initated service ' + ','.join(diff_service)
            )

            self._call_backend_app(
                'up', services=diff_service, options_args=options_args
            )
        services_names = self._get_services_names(services=services, all=all)
        print(services_names)
        self._call_backend_app(
            'up', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_stop(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Stop services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'stop', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_stats(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Display a live stream of container(s) resource usage statistics."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'stats', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_unpause(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Unpause services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'unpause', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_up(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Create and start containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'up', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_version(
        self,
        options: str = '',
    ) -> None:
        """Show the podman-compose version information."""
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'version', services=[], options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_wait(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """
        Block until the first service container stops.

        Note: This is an experimental feature.
        """
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'wait', services=services_names, options_args=options_args
        )
